# coding: utf-8

from __future__ import unicode_literals

"""
This module implements input and output processing from QChem.
"""

import copy
import re
import numpy as np
from string import Template

import six

from monty.io import zopen
from pymatgen.core.operations import SymmOp
from pymatgen.core.structure import Molecule
from pymatgen.core.units import Energy, FloatWithUnit
from pymatgen.serializers.json_coders import PMGSONable
from pymatgen.util.coord_utils import get_angle
from six.moves import map, zip

__author__ = "Xiaohui Qu"
__copyright__ = "Copyright 2013, The Electrolyte Genome Project"
__version__ = "0.1"
__maintainer__ = "Xiaohui Qu"
__email__ = "xhqu1981@gmail.com"
__date__ = "11/4/13"


class QcTask(PMGSONable):
    """
    An object representing a QChem input file.

    Args:
        molecule: The input molecule. If it is None of string "read",
            QChem will read geometry from checkpoint file. If it is a
            Molecule object, QcInput will convert it into Cartesian
            coordinates. Valid values: pymatgen Molecule object, "read", None
            Defaults to None.
        charge (int): Charge of the molecule. If None, charge on molecule is
            used. Defaults to None.
        spin_multiplicity (int): Spin multiplicity of molecule. Defaults to
            None, which means that the spin multiplicity is set to 1 if the
            molecule has no unpaired electrons and to 2 if there are
            unpaired electrons.
        jobtype (str): The type the QChem job. "SP" for Single Point Energy,
            "opt" for geometry optimization, "freq" for
            vibrational frequency.
        title (str): Comments for the job. Defaults to None. Which means the
            $comment section will be discarded.
        exchange (str): The exchange methods of the theory. Examples including:
            "B" (in pure BLYP), "PW91", "PBE", "TPSS".
            Defaults to "HF".
            This parameter can also be common names of hybrid
            functionals, such as B3LYP, TPSSh, XYGJOS. In such cases,
            the correlation parameter should be left as None.
        correlation (str): The correlation level of the theory. Example
            including: "MP2", "RI-MP2", "CCSD(T)", "LYP", "PBE", "TPSS"
            Defaults to None.
        basis_set (str/dict): The basis set.
            If it is a dict, each element can use different basis set.
        aux_basis_set (str/dict): Auxiliary basis set. For methods,
            like RI-MP2, XYG3, OXYJ-OS, auxiliary basis set is required.
            If it is a dict, each element can use different auxiliary
            basis set.
        ecp: Effective core potential (ECP) to be used.
            If it is a dict, each element can use different ECP.
        rem_params (dict): The parameters supposed to write in the $rem
            section. Dict of key/value pairs.
            Example: {"scf_algorithm": "diis_gdm", "scf_max_cycles": 100}
        optional_params (dict): The parameter for keywords other than $rem
            section. Dict of key/value pairs.
            Example: {"basis": {"Li": "cc-PVTZ", "B": "aug-cc-PVTZ",
            "F": "aug-cc-PVTZ"} "ecp": {"Cd": "srsc", "Br": "srlc"}}
        ghost_atom (list): List of ghost atoms indices. Indices start from 0.
            The ghost atom will be represented in of the form of @element_symmbol
    """

    optional_keywords_list = {"basis", "ecp", "empirical_dispersion",
                              "external_charges", "force_field_params",
                              "intracule", "isotopes", "aux_basis",
                              "localized_diabatization", "multipole_field",
                              "nbo", "occupied", "swap_occupied_virtual", "opt",
                              "pcm", "pcm_solvent", "plots", "qm_atoms", "svp",
                              "svpirf", "van_der_waals", "xc_functional",
                              "cdft", "efp_fragments", "efp_params"}
    alternative_keys = {"job_type": "jobtype",
                        "symmetry_ignore": "sym_ignore",
                        "scf_max_cycles": "max_scf_cycles"}
    alternative_values = {"optimization": "opt",
                          "frequency": "freq"}
    zmat_patt = re.compile("^(\w+)*([\s,]+(\w+)[\s,]+(\w+))*[\-\.\s,\w]*$")
    xyz_patt = re.compile("^(\w+)[\s,]+([\d\.eE\-]+)[\s,]+([\d\.eE\-]+)[\s,]+"
                          "([\d\.eE\-]+)[\-\.\s,\w.]*$")

    def __init__(self, molecule=None, charge=None, spin_multiplicity=None,
                 jobtype='SP', title=None, exchange="HF", correlation=None,
                 basis_set="6-31+G*", aux_basis_set=None, ecp=None,
                 rem_params=None, optional_params=None, ghost_atoms=None):
        self.mol = copy.deepcopy(molecule) if molecule else "read"
        self.charge = charge
        self.spin_multiplicity = spin_multiplicity
        if isinstance(self.mol, six.string_types):
            self.mol = self.mol.lower()
            if self.mol != "read":
                raise ValueError('The only accept text value for mol is "read"')
        elif isinstance(self.mol, list):
            for m in self.mol:
                if not isinstance(m, Molecule):
                    raise ValueError("In case of type list, every element of mol must be a pymatgen Molecule")
            if self.charge is None or self.spin_multiplicity is None:
                raise ValueError("For fragments molecule section input, charge and spin_multiplicity "
                                 "must be specificed")
            total_charge = sum([m.charge for m in self.mol])
            total_unpaired_electron = sum([m.spin_multiplicity-1 for m in self.mol])
            if total_charge != self.charge:
                raise ValueError("The charge of the molecule doesn't equal to the sum of the fragment charges")
            if total_unpaired_electron % 2 != (self.spin_multiplicity - 1) % 2:
                raise ValueError("Spin multiplicity of molecule and fragments doesn't match")
        elif isinstance(self.mol, Molecule):
            self.charge = charge if charge is not None else self.mol.charge
            ghost_nelectrons = 0
            if ghost_atoms:
                for i in ghost_atoms:
                    site = self.mol.sites[i]
                    for sp, amt in site.species_and_occu.items():
                        ghost_nelectrons += sp.Z * amt
            nelectrons = self.mol.charge + self.mol.nelectrons - ghost_nelectrons - self.charge
            if spin_multiplicity is not None:
                self.spin_multiplicity = spin_multiplicity
                if (nelectrons + spin_multiplicity) % 2 != 1:
                    raise ValueError("Charge of {} and spin multiplicity of {} "
                                     "is not possible for this molecule"
                                     .format(self.charge, spin_multiplicity))
            else:
                self.spin_multiplicity = 1 if nelectrons % 2 == 0 else 2
        else:
            raise ValueError("The molecule must be a pymatgen Molecule "
                             "object or read/None or list of pymatgen Molecule")
        if (self.charge is None) != (self.spin_multiplicity is None):
            raise ValueError("spin multiplicity must be set together")
        if self.charge is not None and isinstance(self.mol, Molecule) and not ghost_atoms:
            self.mol.set_charge_and_spin(self.charge, self.spin_multiplicity)
        self.params = dict()
        if title is not None:
            self.params["comment"] = title
        if "rem" not in self.params:
            self.params["rem"] = dict()
        self.params["rem"]["exchange"] = exchange.lower()
        available_jobtypes = {"sp", "opt", "ts", "freq", "force", "rpath",
                              "nmr", "bsse", "eda", "pes_scan", "fsm", "aimd",
                              "pimc", "makeefp"}
        jt = jobtype.lower()
        if jt in self.alternative_values:
            jt = self.alternative_values[jt]
        if jt not in available_jobtypes:
            raise ValueError("Job type " + jobtype + " is not supported yet")
        self.params["rem"]["jobtype"] = jobtype.lower()
        if correlation is not None:
            self.params["rem"]["correlation"] = correlation.lower()
        if rem_params is not None:
            for k, v in rem_params.items():
                k = k.lower()
                if k in self.alternative_keys:
                    k = self.alternative_keys[k]
                if isinstance(v, six.string_types):
                    v = str(v).lower()
                    if v in self.alternative_values:
                        # noinspection PyTypeChecker
                        v = self.alternative_values[v]
                    self.params["rem"][k] = v
                elif isinstance(v, int) or isinstance(v, float):
                    self.params["rem"][k] = v
                else:
                    raise ValueError("The value in $rem can only be Integer "
                                     "or string")
        if optional_params:
            op_key = set([k.lower() for k in optional_params.keys()])
            if len(op_key - self.optional_keywords_list) > 0:
                invalid_keys = op_key - self.optional_keywords_list
                raise ValueError(','.join(['$' + k for k in invalid_keys]) +
                                 'is not a valid optional section')
            self.params.update(optional_params)

        self.set_basis_set(basis_set)

        if aux_basis_set is None:
            if self._aux_basis_required():
                if isinstance(self.params["rem"]["basis"], six.string_types):
                    if self.params["rem"]["basis"].startswith("6-31+g"):
                        self.set_auxiliary_basis_set("rimp2-aug-cc-pvdz")
                    elif self.params["rem"]["basis"].startswith("6-311+g"):
                        self.set_auxiliary_basis_set("rimp2-aug-cc-pvtz")
                if "aux_basis" not in self.params["rem"]:
                    raise ValueError("Auxiliary basis set is missing")
        else:
            self.set_auxiliary_basis_set(aux_basis_set)

        if ecp:
            self.set_ecp(ecp)
        self.ghost_atoms = ghost_atoms
        if self.ghost_atoms:
            if not isinstance(self.ghost_atoms, list):
                raise ValueError("ghost_atoms must be a list of integers")
            for atom in self.ghost_atoms:
                if not isinstance(atom, int):
                    raise ValueError("Each element of ghost atom list must an integer")

    def _aux_basis_required(self):
        if self.params["rem"]["exchange"] in ['xygjos', 'xyg3', 'lxygjos']:
            return True
        if 'correlation' in self.params["rem"]:
            if self.params["rem"]["correlation"].startswith("ri"):
                return True

    def set_basis_set(self, basis_set):
        if isinstance(basis_set, six.string_types):
            self.params["rem"]["basis"] = str(basis_set).lower()
            if basis_set.lower() not in ["gen", "mixed"]:
                self.params.pop("basis", None)
        elif isinstance(basis_set, dict):
            self.params["rem"]["basis"] = "gen"
            bs = dict()
            for element, basis in basis_set.items():
                bs[element.strip().capitalize()] = basis.lower()
            self.params["basis"] = bs
            if self.mol:
                mol_elements = set([site.species_string for site
                                    in self.mol.sites])
                basis_elements = set(self.params["basis"].keys())
                if len(mol_elements - basis_elements) > 0:
                    raise ValueError("The basis set for elements " +
                                     ", ".join(
                                         list(mol_elements - basis_elements)) +
                                     " is missing")
                if len(basis_elements - mol_elements) > 0:
                    raise ValueError("Basis set error: the molecule "
                                     "doesn't contain element " +
                                     ", ".join(basis_elements - mol_elements))
        elif isinstance(basis_set, list):
            self.params["rem"]["basis"] = "mixed"
            bs = [(a[0].capitalize(), a[1].lower()) for a in basis_set]
            self.params["basis"] = bs
            if len(self.mol) != len(basis_set):
                raise ValueError("Must specific a basis set for every atom")
            mol_elements = [site.species_string for site in self.mol.sites]
            basis_elements = [a[0] for a in bs]
            if mol_elements != basis_elements:
                raise ValueError("Elements in molecule and mixed basis set don't match")
        else:
            raise Exception('Can\'t handle type "{}"'.format(type(basis_set)))

    def set_auxiliary_basis_set(self, aux_basis_set):
        if isinstance(aux_basis_set, six.string_types):
            self.params["rem"]["aux_basis"] = aux_basis_set.lower()
            if aux_basis_set.lower() not in ["gen", "mixed"]:
                self.params.pop("aux_basis", None)
        elif isinstance(aux_basis_set, dict):
            self.params["rem"]["aux_basis"] = "gen"
            bs = dict()
            for element, basis in aux_basis_set.items():
                bs[element.strip().capitalize()] = basis.lower()
            self.params["aux_basis"] = bs
            if self.mol:
                mol_elements = set([site.species_string for site
                                    in self.mol.sites])
                basis_elements = set(self.params["aux_basis"].keys())
                if len(mol_elements - basis_elements) > 0:
                    raise ValueError("The auxiliary basis set for "
                                     "elements " +
                                     ", ".join(
                                         list(mol_elements - basis_elements)) +
                                     " is missing")
                if len(basis_elements - mol_elements) > 0:
                    raise ValueError("Auxiliary asis set error: the "
                                     "molecule doesn't contain element " +
                                     ", ".join(basis_elements - mol_elements))
        elif isinstance(aux_basis_set, list):
            self.params["rem"]["aux_basis"] = "mixed"
            bs = [(a[0].capitalize(), a[1].lower()) for a in aux_basis_set]
            self.params["aux_basis"] = bs
            if len(self.mol) != len(aux_basis_set):
                raise ValueError("Must specific a auxiliary basis set for every atom")
            mol_elements = [site.species_string for site in self.mol.sites]
            basis_elements = [a[0] for a in bs]
            if mol_elements != basis_elements:
                raise ValueError("Elements in molecule and mixed basis set don't match")
        else:
            raise Exception('Can\'t handle type "{}"'.format(type(aux_basis_set)))

    def set_ecp(self, ecp):
        if isinstance(ecp, six.string_types):
            self.params["rem"]["ecp"] = ecp.lower()
        elif isinstance(ecp, dict):
            self.params["rem"]["ecp"] = "gen"
            potentials = dict()
            for element, p in ecp.items():
                potentials[element.strip().capitalize()] = p.lower()
            self.params["ecp"] = potentials
            if self.mol:
                mol_elements = set([site.species_string for site
                                    in self.mol.sites])
                ecp_elements = set(self.params["ecp"].keys())
                if len(ecp_elements - mol_elements) > 0:
                    raise ValueError("ECP error: the molecule "
                                     "doesn't contain element " +
                                     ", ".join(ecp_elements - mol_elements))

    @property
    def molecule(self):
        return self.mol

    def set_memory(self, total=None, static=None):
        """
        Set the maxium allowed memory.

        Args:
            total: The total memory. Integer. Unit: MBytes. If set to None,
                this parameter will be neglected.
            static: The static memory. Integer. Unit MBytes. If set to None,
                this parameterwill be neglected.
        """
        if total:
            self.params["rem"]["mem_total"] = total
        if static:
            self.params["rem"]["mem_static"] = static

    def set_max_num_of_scratch_files(self, num=16):
        """
        In QChem, the size of a single scratch is limited 2GB. By default,
        the max number of scratich is 16, which is cooresponding to 32GB
        scratch space. If you want to use more scratch disk space, you need
        to increase the number of scratch files:

        Args:
            num: The max number of the scratch files. (Integer)
        """
        self.params["rem"]["max_sub_file_num"] = num

    def set_scf_algorithm_and_iterations(self, algorithm="diis",
                                         iterations=50):
        """
        Set algorithm used for converging SCF and max number of SCF iterations.

        Args:
            algorithm: The algorithm used for converging SCF. (str)
            iterations: The max number of SCF iterations. (Integer)
        """
        available_algorithms = {"diis", "dm", "diis_dm", "diis_gdm", "gdm",
                                "rca", "rca_diis", "roothaan"}
        if algorithm.lower() not in available_algorithms:
            raise ValueError("Algorithm " + algorithm +
                             " is not available in QChem")
        self.params["rem"]["scf_algorithm"] = algorithm.lower()
        self.params["rem"]["max_scf_cycles"] = iterations

    def set_scf_convergence_threshold(self, exponent=8):
        """
        SCF is considered converged when the wavefunction error is less than
        10**(-exponent).
        In QChem, the default values are:
        5	For single point energy calculations.
        7	For geometry optimizations and vibrational analysis.
        8	For SSG calculations

        Args:
            exponent: The exponent of the threshold. (Integer)
        """
        self.params["rem"]["scf_convergence"] = exponent

    def set_integral_threshold(self, thresh=12):
        """
        Cutoff for neglect of two electron integrals. 10−THRESH (THRESH <= 14).
        In QChem, the default values are:
        8	For single point energies.
        10	For optimizations and frequency calculations.
        14	For coupled-cluster calculations.

        Args:
            thresh: The exponent of the threshold. (Integer)
        """
        self.params["rem"]["thresh"] = thresh

    def set_dft_grid(self, radical_points=128, angular_points=302,
                     grid_type="Lebedev"):
        """
        Set the grid for DFT numerical integrations.

        Args:
            radical_points: Radical points. (Integer)
            angular_points: Angular points. (Integer)
            grid_type: The type of of the grid. There are two standard grids:
                SG-1 and SG-0. The other two supported grids are "Lebedev" and
                "Gauss-Legendre"
        """
        available_lebedev_angular_points = {6, 18, 26, 38, 50, 74, 86, 110, 146,
                                            170, 194, 230, 266, 302, 350, 434,
                                            590, 770, 974, 1202, 1454, 1730,
                                            2030, 2354, 2702, 3074, 3470, 3890,
                                            4334, 4802, 5294}
        if grid_type.lower() == "sg-0":
            self.params["rem"]["xc_grid"] = 0
        elif grid_type.lower() == "sg-1":
            self.params["rem"]["xc_grid"] = 1
        elif grid_type.lower() == "lebedev":
            if angular_points not in available_lebedev_angular_points:
                raise ValueError(str(angular_points) + " is not a valid "
                                 "Lebedev angular points number")
            self.params["rem"]["xc_grid"] = "{rp:06d}{ap:06d}".format(
                rp=radical_points, ap=angular_points)
        elif grid_type.lower() == "gauss-legendre":
            self.params["rem"]["xc_grid"] = "-{rp:06d}{ap:06d}".format(
                rp=radical_points, ap=angular_points)
        else:
            raise ValueError("Grid type " + grid_type + " is not supported "
                                                        "currently")

    def set_scf_initial_guess(self, guess="SAD"):
        """
        Set initial guess method to be used for SCF

        Args:
            guess: The initial guess method. (str)
        """
        availabel_guesses = {"core", "sad", "gwh", "read", "fragmo"}
        if guess.lower() not in availabel_guesses:
            raise ValueError("The guess method " + guess + " is not supported "
                                                           "yet")
        self.params["rem"]["scf_guess"] = guess.lower()

    def set_geom_max_iterations(self, iterations):
        """
        Set the max iterations of geometry optimization.

        Args:
            iterations: the maximum iterations of geometry optimization.
            (Integer)
        """
        self.params["rem"]["geom_opt_max_cycles"] = iterations

    def set_geom_opt_coords_type(self, coords_type="internal_switch"):
        """
        Set the coordinates system used in geometry optimization.
        "cartesian"       --- always cartesian coordinates.
        "internal"        --- always internal coordinates.
        "internal-switch" --- try internal coordinates first, if fails, switch
        to cartesian coordinates.
        "z-matrix"        --- always z-matrix coordinates.
        "z-matrix-switch" --- try z-matrix first, if fails, switch to
        cartesian coordinates.

        Args:
            coords_type: The type of the coordinates. (str)
        """
        coords_map = {"cartesian": 0, "internal": 1, "internal-switch": -1,
                      "z-matrix": 2, "z-matrix-switch": -2}
        if coords_type.lower() not in set(coords_map.keys()):
            raise ValueError("Coodinate system " + coords_type + " is not "
                             "supported yet")
        else:
            self.params["rem"]["geom_opt_coords"] = \
                coords_map[coords_type.lower()]

    def scale_geom_opt_threshold(self, gradient=0.1, displacement=0.1,
                                 energy=0.1):
        """
        Adjust the convergence criteria of geometry optimization.

        Args:
            gradient: the scale factor for gradient criteria. If less than
                1.0, you are tightening the threshold. The base value is
                300 × 10E−6
            displacement: the scale factor for atomic displacement. If less
                then 1.0, you are tightening the threshold. The base value is
                1200 × 10E−6
            energy: the scale factor for energy change between successive
                iterations. If less than 1.0, you are tightening the
                threshold. The base value is 100 × 10E−8.
        """
        if gradient < 1.0/(300-1) or displacement < 1.0/(1200-1) or \
                energy < 1.0/(100-1):
            raise ValueError("The geometry optimization convergence criteria "
                             "is too tight")
        self.params["rem"]["geom_opt_tol_gradient"] = int(gradient * 300)
        self.params["rem"]["geom_opt_tol_displacement"] = int(displacement *
                                                              1200)
        self.params["rem"]["geom_opt_tol_energy"] = int(energy * 100)

    def set_geom_opt_use_gdiis(self, subspace_size=None):
        """
        Use GDIIS algorithm in geometry optimization.

        Args:
            subspace_size: The size of the DIIS subsapce. None for default
                value. The default value is min(NDEG, NATOMS, 4) NDEG = number
                of moleculardegrees of freedom.
        """
        subspace_size = subspace_size if subspace_size is not None else -1
        self.params["rem"]["geom_opt_max_diis"] = subspace_size

    def disable_symmetry(self):
        """
        Turn the symmetry off.
        """
        self.params["rem"]["sym_ignore"] = True
        self.params["rem"]["symmetry"] = False

    def use_cosmo(self, dielectric_constant=78.4):
        """
        Set the solvent model to COSMO.

        Args:
            dielectric_constant: the dielectric constant for the solvent.
        """
        self.params["rem"]["solvent_method"] = "cosmo"
        self.params["rem"]["solvent_dielectric"] = dielectric_constant

    def use_pcm(self, pcm_params=None, solvent_params=None,
                radii_force_field=None):
        """
        Set the solvent model to PCM. Default parameters are trying to comply to
        gaussian default value

        Args:
            pcm_params (dict): The parameters of "$pcm" section.
            solvent_params (dict): The parameters of "pcm_solvent" section
            radii_force_field (str): The force fied used to set the solute
                radii. Default to UFF.
        """
        self.params["pcm"] = dict()
        self.params["pcm_solvent"] = dict()
        default_pcm_params = {"Theory": "SSVPE",
                              "vdwScale": 1.1,
                              "Radii": "UFF"}
        if not solvent_params:
            solvent_params = {"Dielectric": 78.3553}
        if pcm_params:
            for k, v in pcm_params.items():
                self.params["pcm"][k.lower()] = v.lower() \
                    if isinstance(v, six.string_types) else v

        for k, v in default_pcm_params.items():
            if k.lower() not in self.params["pcm"].keys():
                self.params["pcm"][k.lower()] = v.lower() \
                    if isinstance(v, six.string_types) else v
        for k, v in solvent_params.items():
            self.params["pcm_solvent"][k.lower()] = v.lower() \
                if isinstance(v, six.string_types) else copy.deepcopy(v)
        self.params["rem"]["solvent_method"] = "pcm"
        if radii_force_field:
            self.params["pcm"]["radii"] = "bondi"
            self.params["rem"]["force_fied"] = radii_force_field.lower()

    def __str__(self):
        sections = ["comment", "molecule", "rem"] + \
            sorted(list(self.optional_keywords_list))
        lines = []
        for sec in sections:
            if sec in self.params or sec == "molecule":
                foramt_sec = self.__getattribute__("_format_" + sec)
                lines.append("$" + sec)
                lines.extend(foramt_sec())
                lines.append("$end")
                lines.append('\n')
        return '\n'.join(lines)

    def _format_comment(self):
        lines = [' ' + self.params["comment"].strip()]
        return lines

    def _format_molecule(self):
        lines = []

        def inner_format_mol(m2, index_base):
            mol_lines = []
            for i, site in enumerate(m2.sites):
                ghost = "@" if self.ghost_atoms \
                    and i + index_base in self.ghost_atoms else ""
                atom = "{ghost:s}{element:s}".format(ghost=ghost,
                                                     element=site.species_string)
                mol_lines.append(" {atom:<4} {x:>17.8f} {y:>17.8f} "
                                 "{z:>17.8f}".format(atom=atom, x=site.x,
                                                     y=site.y, z=site.z))
            return mol_lines

        if self.charge is not None:
            lines.append(" {charge:d}  {multi:d}".format(charge=self
                         .charge, multi=self.spin_multiplicity))
        if isinstance(self.mol, six.string_types) and self.mol == "read":
            lines.append(" read")
        elif isinstance(self.mol, list):
            starting_index = 0
            for m in self.mol:
                lines.append("--")
                lines.append(" {charge:d}  {multi:d}".format(
                    charge=m.charge, multi=m.spin_multiplicity))
                lines.extend(inner_format_mol(m, starting_index))
                starting_index += len(m)
        else:
            lines.extend(inner_format_mol(self.mol, 0))
        return lines

    def _format_rem(self):
        rem_format_template = Template("  {name:>$name_width} = "
                                       "{value}")
        name_width = 0
        for name, value in self.params["rem"].items():
            if len(name) > name_width:
                name_width = len(name)
        rem = rem_format_template.substitute(name_width=name_width)
        lines = []
        all_keys = set(self.params["rem"].keys())
        priority_keys = ["jobtype", "exchange", "basis"]
        additional_keys = all_keys - set(priority_keys)
        ordered_keys = priority_keys + sorted(list(additional_keys))
        for name in ordered_keys:
            value = self.params["rem"][name]
            lines.append(rem.format(name=name, value=value))
        return lines

    def _format_basis(self):
        lines = []
        if isinstance(self.params["basis"], dict):
            for element in sorted(self.params["basis"].keys()):
                basis = self.params["basis"][element]
                lines.append(" " + element)
                lines.append(" " + basis)
                lines.append(" ****")
        elif isinstance(self.params["basis"], list):
            for i, (element, bs) in enumerate(self.params["basis"]):
                lines.append(" {element:2s} {number:3d}".format(element=element, number=i+1))
                lines.append(" {}".format(bs))
                lines.append(" ****")
        return lines

    def _format_aux_basis(self):
        lines = []
        if isinstance(self.params["aux_basis"], dict):
            for element in sorted(self.params["aux_basis"].keys()):
                basis = self.params["aux_basis"][element]
                lines.append(" " + element)
                lines.append(" " + basis)
                lines.append(" ****")
        else:
            for i, (element, bs) in enumerate(self.params["aux_basis"]):
                lines.append(" {element:2s} {number:3d}".format(element=element, number=i+1))
                lines.append(" {}".format(bs))
                lines.append(" ****")
        return lines

    def _format_ecp(self):
        lines = []
        for element in sorted(self.params["ecp"].keys()):
            ecp = self.params["ecp"][element]
            lines.append(" " + element)
            lines.append(" " + ecp)
            lines.append(" ****")
        return lines

    def _format_pcm(self):
        pcm_format_template = Template("  {name:>$name_width}   "
                                       "{value}")
        name_width = 0
        for name in self.params["pcm"].keys():
            if len(name) > name_width:
                name_width = len(name)
        rem = pcm_format_template.substitute(name_width=name_width)
        lines = []
        for name in sorted(self.params["pcm"].keys()):
            value = self.params["pcm"][name]
            lines.append(rem.format(name=name, value=value))
        return lines

    def _format_pcm_solvent(self):
        pp_format_template = Template("  {name:>$name_width}   "
                                      "{value}")
        name_width = 0
        for name in self.params["pcm_solvent"].keys():
            if len(name) > name_width:
                name_width = len(name)
        rem = pp_format_template.substitute(name_width=name_width)
        lines = []
        all_keys = set(self.params["pcm_solvent"].keys())
        priority_keys = []
        for k in ["dielectric", "nonels", "nsolventatoms", "solventatom"]:
            if k in all_keys:
                priority_keys.append(k)
        additional_keys = all_keys - set(priority_keys)
        ordered_keys = priority_keys + sorted(list(additional_keys))
        for name in ordered_keys:
            value = self.params["pcm_solvent"][name]
            if name == "solventatom":
                for v in copy.deepcopy(value):
                    value = "{:<4d} {:<4d} {:<4d} {:4.2f}".format(*v)
                    lines.append(rem.format(name=name, value=value))
                continue
            lines.append(rem.format(name=name, value=value))
        return lines

    def as_dict(self):
        if isinstance(self.mol, six.string_types):
            mol_dict = self.mol
        elif isinstance(self.mol, Molecule):
            mol_dict = self.mol.as_dict()
        elif isinstance(self.mol, list):
            mol_dict = [m.as_dict() for m in self.mol]
        else:
            raise ValueError('Unknow molecule type "{}"'.format(type(self.mol)))
        d = {"@module": self.__class__.__module__,
             "@class": self.__class__.__name__,
             "molecule": mol_dict,
             "charge": self.charge,
             "spin_multiplicity": self.spin_multiplicity,
             "params": self.params}
        if self.ghost_atoms:
            d["ghost_atoms"] = self.ghost_atoms
        return d

    @classmethod
    def from_dict(cls, d):
        if d["molecule"] == "read":
            mol = "read"
        elif isinstance(d["molecule"], dict):
            mol = Molecule.from_dict(d["molecule"])
        elif isinstance(d["molecule"], list):
            mol = [Molecule.from_dict(m) for m in d["molecule"]]
        else:
            raise ValueError('Unknow molecule type "{}"'.format(type(d["molecule"])))
        jobtype = d["params"]["rem"]["jobtype"]
        title = d["params"].get("comment", None)
        exchange = d["params"]["rem"]["exchange"]
        correlation = d["params"]["rem"].get("correlation", None)
        basis_set = d["params"]["rem"]["basis"]
        aux_basis_set = d["params"]["rem"].get("aux_basis", None)
        ecp = d["params"]["rem"].get("ecp", None)
        ghost_atoms = d.get("ghost_atoms", None)
        optional_params = None
        op_keys = set(d["params"].keys()) - {"comment", "rem"}
        if len(op_keys) > 0:
            optional_params = dict()
            for k in op_keys:
                optional_params[k] = d["params"][k]
        return QcTask(molecule=mol, charge=d["charge"],
                      spin_multiplicity=d["spin_multiplicity"],
                      jobtype=jobtype, title=title,
                      exchange=exchange, correlation=correlation,
                      basis_set=basis_set, aux_basis_set=aux_basis_set,
                      ecp=ecp, rem_params=d["params"]["rem"],
                      optional_params=optional_params,
                      ghost_atoms=ghost_atoms)

    def write_file(self, filename):
        with zopen(filename, "wt") as f:
            f.write(self.__str__())

    @classmethod
    def from_file(cls, filename):
        with zopen(filename, "rt") as f:
            return cls.from_string(f.read())

    @classmethod
    def from_string(cls, contents):
        """
        Creates QcInput from a string.

        Args:
            contents: String representing a QChem input file.

        Returns:
            QcInput object
        """
        mol = None
        charge = None
        spin_multiplicity = None
        params = dict()
        lines = contents.split('\n')
        parse_section = False
        section_name = None
        section_text = []
        ghost_atoms = None
        for line_num, line in enumerate(lines):
            l = line.strip().lower()

            if len(l) == 0:
                continue
            if (not parse_section) and (l == "$end" or not l.startswith("$")):
                raise ValueError("Format error, parsing failed")
            if parse_section and l != "$end":
                section_text.append(line)
            if l.startswith("$") and not parse_section:
                parse_section = True
                section_name = l[1:]
                available_sections = ["comment", "molecule", "rem"] + \
                    sorted(list(cls.optional_keywords_list))
                if section_name not in available_sections:
                    raise ValueError("Unrecognized keyword " + line.strip() +
                                     " at line " + str(line_num))
                if section_name in params:
                    raise ValueError("duplicated keyword " + line.strip() +
                                     "at line " + str(line_num))
            if parse_section and l == "$end":
                func_name = "_parse_" + section_name
                if func_name not in QcTask.__dict__:
                    raise Exception(func_name + " is not implemented yet, "
                                    "please implement it")
                parse_func = QcTask.__dict__[func_name].__get__(None, QcTask)
                if section_name == "molecule":
                    mol, charge, spin_multiplicity, ghost_atoms = parse_func(section_text)
                else:
                    d = parse_func(section_text)
                    params[section_name] = d
                parse_section = False
                section_name = None
                section_text = []
        if parse_section:
            raise ValueError("Format error. " + section_name + " is not "
                             "terminated")
        jobtype = params["rem"]["jobtype"]
        title = params.get("comment", None)
        exchange = params["rem"].get("exchange", "hf")
        correlation = params["rem"].get("correlation", None)
        basis_set = params["rem"]["basis"]
        aux_basis_set = params["rem"].get("aux_basis", None)
        ecp = params["rem"].get("ecp", None)
        optional_params = None
        op_keys = set(params.keys()) - {"comment", "rem"}
        if len(op_keys) > 0:
            optional_params = dict()
            for k in op_keys:
                optional_params[k] = params[k]
        return QcTask(molecule=mol, charge=charge,
                      spin_multiplicity=spin_multiplicity,
                      jobtype=jobtype, title=title,
                      exchange=exchange, correlation=correlation,
                      basis_set=basis_set, aux_basis_set=aux_basis_set,
                      ecp=ecp, rem_params=params["rem"],
                      optional_params=optional_params,
                      ghost_atoms=ghost_atoms)

    @classmethod
    def _parse_comment(cls, contents):
        return '\n'.join(contents).strip()

    @classmethod
    def _parse_coords(cls, coord_lines):
        """
        Helper method to parse coordinates. Copied from GaussianInput class.
        """
        paras = {}
        var_pattern = re.compile("^([A-Za-z]+\S*)[\s=,]+([\d\-\.]+)$")
        for l in coord_lines:
            m = var_pattern.match(l.strip())
            if m:
                paras[m.group(1)] = float(m.group(2))

        species = []
        coords = []
        # Stores whether a Zmatrix format is detected. Once a zmatrix format
        # is detected, it is assumed for the remaining of the parsing.
        zmode = False
        for l in coord_lines:
            l = l.strip()
            if not l:
                break
            if (not zmode) and cls.xyz_patt.match(l):
                m = cls.xyz_patt.match(l)
                species.append(m.group(1))
                toks = re.split("[,\s]+", l.strip())
                if len(toks) > 4:
                    coords.append(list(map(float, toks[2:5])))
                else:
                    coords.append(list(map(float, toks[1:4])))
            elif cls.zmat_patt.match(l):
                zmode = True
                toks = re.split("[,\s]+", l.strip())
                species.append(toks[0])
                toks.pop(0)
                if len(toks) == 0:
                    coords.append(np.array([0.0, 0.0, 0.0]))
                else:
                    nn = []
                    parameters = []
                    while len(toks) > 1:
                        ind = toks.pop(0)
                        data = toks.pop(0)
                        try:
                            nn.append(int(ind))
                        except ValueError:
                            nn.append(species.index(ind) + 1)
                        try:
                            val = float(data)
                            parameters.append(val)
                        except ValueError:
                            if data.startswith("-"):
                                parameters.append(-paras[data[1:]])
                            else:
                                parameters.append(paras[data])
                    if len(nn) == 1:
                        coords.append(np.array(
                            [0.0, 0.0, float(parameters[0])]))
                    elif len(nn) == 2:
                        coords1 = coords[nn[0] - 1]
                        coords2 = coords[nn[1] - 1]
                        bl = parameters[0]
                        angle = parameters[1]
                        axis = [0, 1, 0]
                        op = SymmOp.from_origin_axis_angle(coords1, axis,
                                                           angle, False)
                        coord = op.operate(coords2)
                        vec = coord - coords1
                        coord = vec * bl / np.linalg.norm(vec) + coords1
                        coords.append(coord)
                    elif len(nn) == 3:
                        coords1 = coords[nn[0] - 1]
                        coords2 = coords[nn[1] - 1]
                        coords3 = coords[nn[2] - 1]
                        bl = parameters[0]
                        angle = parameters[1]
                        dih = parameters[2]
                        v1 = coords3 - coords2
                        v2 = coords1 - coords2
                        axis = np.cross(v1, v2)
                        op = SymmOp.from_origin_axis_angle(
                            coords1, axis, angle, False)
                        coord = op.operate(coords2)
                        v1 = coord - coords1
                        v2 = coords1 - coords2
                        v3 = np.cross(v1, v2)
                        adj = get_angle(v3, axis)
                        axis = coords1 - coords2
                        op = SymmOp.from_origin_axis_angle(
                            coords1, axis, dih - adj, False)
                        coord = op.operate(coord)
                        vec = coord - coords1
                        coord = vec * bl / np.linalg.norm(vec) + coords1
                        coords.append(coord)

        def parse_species(sp_str):
            """
            The species specification can take many forms. E.g.,
            simple integers representing atomic numbers ("8"),
            actual species string ("C") or a labelled species ("C1").
            Sometimes, the species string is also not properly capitalized,
            e.g, ("c1"). This method should take care of these known formats.
            """
            try:
                return int(sp_str)
            except ValueError:
                sp = re.sub("\d", "", sp_str)
                return sp.capitalize()

        species = list(map(parse_species, species))

        return Molecule(species, coords)

    @classmethod
    def _parse_molecule(cls, contents):
        def parse_ghost_indices(coord_text_lines):
            no_ghost_text = [l.replace("@", "") for l in coord_text_lines]
            ghosts = []
            for index, l in enumerate(coord_text_lines):
                l = l.strip()
                if not l:
                    break
                if "@" in l:
                    ghosts.append(index)
            return ghosts, no_ghost_text

        text = copy.deepcopy(contents[:2])
        charge_multi_pattern = re.compile('\s*(?P<charge>'
                                          '[-+]?\d+)\s+(?P<multi>\d+)')
        line = text.pop(0)
        m = charge_multi_pattern.match(line)
        if m:
            charge = int(m.group("charge"))
            spin_multiplicity = int(m.group("multi"))
            line = text.pop(0)
        else:
            charge = None
            spin_multiplicity = None
        if line.strip().lower() == "read":
            return "read", charge, spin_multiplicity, None
        elif charge is None or spin_multiplicity is None:
            raise ValueError("Charge or spin multiplicity is not found")
        else:
            if contents[1].strip()[0:2] == "--":
                chunks = "\n".join(contents[2:]).split("--\n")
                mol = []
                ghost_atoms = []
                starting_index = 0
                for chunk in chunks:
                    frag_contents = chunk.split("\n")
                    m = charge_multi_pattern.match(frag_contents[0])
                    if m:
                        fragment_charge = int(m.group("charge"))
                        fragment_spin_multiplicity = int(m.group("multi"))
                    else:
                        raise Exception("charge and spin multiplicity must be specified for each fragment")
                    gh, coord_lines = parse_ghost_indices(frag_contents[1:])
                    fragment = cls._parse_coords(coord_lines)
                    fragment.set_charge_and_spin(fragment_charge, fragment_spin_multiplicity)
                    mol.append(fragment)
                    ghost_atoms.extend([i+starting_index for i in gh])
                    starting_index += len(fragment)
            else:
                ghost_atoms, coord_lines = parse_ghost_indices(contents[1:])
                mol = cls._parse_coords(coord_lines)
                if len(ghost_atoms) == 0:
                    mol.set_charge_and_spin(charge, spin_multiplicity)
            ghost_atoms = ghost_atoms if len(ghost_atoms) > 0 else None
            return mol, charge, spin_multiplicity, ghost_atoms

    @classmethod
    def _parse_rem(cls, contents):
        d = dict()
        int_pattern = re.compile('^[-+]?\d+$')
        float_pattern = re.compile('^[-+]?\d+\.\d+([eE][-+]?\d+)?$')

        for line in contents:
            tokens = line.strip().replace("=", ' ').split()
            if len(tokens) < 2:
                raise ValueError("Can't parse $rem section, there should be "
                                 "at least two field: key and value!")
            k1, v = tokens[:2]
            k2 = k1.lower()
            if k2 in cls.alternative_keys:
                k2 = cls.alternative_keys[k2]
            if v in cls.alternative_values:
                v = cls.alternative_values
            if k2 == "xc_grid":
                d[k2] = v
            elif v == "True":
                d[k2] = True
            elif v == "False":
                d[k2] = False
            elif int_pattern.match(v):
                d[k2] = int(v)
            elif float_pattern.match(v):
                d[k2] = float(v)
            else:
                d[k2] = v.lower()
        return d

    @classmethod
    def _parse_aux_basis(cls, contents):
        if len(contents) % 3 != 0:
            raise ValueError("Auxiliary basis set section format error")
        chunks = zip(*[iter(contents)]*3)
        t = contents[0].split()
        if len(t) == 2 and int(t[1]) > 0:
            bs = []
            for i, ch in enumerate(chunks):
                element, number = ch[0].split()
                basis = ch[1]
                if int(number) != i+1:
                    raise ValueError("Atom order number doesn't match in $aux_basis section")
                bs.append((element.strip().capitalize(), basis.strip().lower()))
        else:
            bs = dict()
            for ch in chunks:
                element, basis = ch[:2]
                bs[element.strip().capitalize()] = basis.strip().lower()
        return bs

    @classmethod
    def _parse_basis(cls, contents):
        if len(contents) % 3 != 0:
            raise ValueError("Basis set section format error")
        chunks = zip(*[iter(contents)]*3)
        t = contents[0].split()
        if len(t) == 2 and int(t[1]) > 0:
            bs = []
            for i, ch in enumerate(chunks):
                element, number = ch[0].split()
                basis = ch[1]
                if int(number) != i+1:
                    raise ValueError("Atom order number doesn't match in $basis section")
                bs.append((element.strip().capitalize(), basis.strip().lower()))
        else:
            bs = dict()
            for ch in chunks:
                element, basis = ch[:2]
                bs[element.strip().capitalize()] = basis.strip().lower()
        return bs

    @classmethod
    def _parse_ecp(cls, contents):
        if len(contents) % 3 != 0:
            raise ValueError("ECP section format error")
        chunks = zip(*[iter(contents)]*3)
        d = dict()
        for ch in chunks:
            element, ecp = ch[:2]
            d[element.strip().capitalize()] = ecp.strip().lower()
        return d

    @classmethod
    def _parse_pcm(cls, contents):
        d = dict()
        int_pattern = re.compile('^[-+]?\d+$')
        float_pattern = re.compile('^[-+]?\d+\.\d+([eE][-+]?\d+)?$')

        for line in contents:
            tokens = line.strip().replace("=", ' ').split()
            if len(tokens) < 2:
                raise ValueError("Can't parse $pcm section, there should be "
                                 "at least two field: key and value!")
            k1, v = tokens[:2]
            k2 = k1.lower()
            if k2 in cls.alternative_keys:
                k2 = cls.alternative_keys[k2]
            if v in cls.alternative_values:
                v = cls.alternative_values
            if v == "True":
                d[k2] = True
            elif v == "False":
                d[k2] = False
            elif int_pattern.match(v):
                d[k2] = int(v)
            elif float_pattern.match(v):
                d[k2] = float(v)
            else:
                d[k2] = v.lower()
        return d

    @classmethod
    def _parse_pcm_solvent(cls, contents):
        d = dict()
        int_pattern = re.compile('^[-+]?\d+$')
        float_pattern = re.compile('^[-+]?\d+\.\d+([eE][-+]?\d+)?$')

        for line in contents:
            tokens = line.strip().replace("=", ' ').split()
            if len(tokens) < 2:
                raise ValueError("Can't parse $pcm_solvent section, "
                                 "there should be at least two field: "
                                 "key and value!")
            k1, v = tokens[:2]
            k2 = k1.lower()
            if k2 in cls.alternative_keys:
                k2 = cls.alternative_keys[k2]
            if v in cls.alternative_values:
                v = cls.alternative_values
            if k2 == "solventatom":
                v = [int(i) for i in tokens[1:4]]
                # noinspection PyTypeChecker
                v.append(float(tokens[4]))
                if k2 not in d:
                    d[k2] = [v]
                else:
                    d[k2].append(v)
            elif v == "True":
                d[k2] = True
            elif v == "False":
                d[k2] = False
            elif int_pattern.match(v):
                d[k2] = int(v)
            elif float_pattern.match(v):
                d[k2] = float(v)
            else:
                d[k2] = v.lower()
        return d


class QcInput(PMGSONable):
    """
    An object representing a multiple step QChem input file.

    Args:
        jobs: The QChem jobs (List of QcInput object)
    """

    def __init__(self, jobs):
        jobs = jobs if isinstance(jobs, list) else [jobs]
        for j in jobs:
            if not isinstance(j, QcTask):
                raise ValueError("jobs must be a list QcInput object")
            self.jobs = jobs

    def __str__(self):
        return "\n@@@\n\n\n".join([str(j) for j in self.jobs])

    def write_file(self, filename):
        with zopen(filename, "wt") as f:
            f.write(self.__str__())

    def as_dict(self):
        return {"@module": self.__class__.__module__,
                "@class": self.__class__.__name__,
                "jobs": [j.as_dict() for j in self.jobs]}

    @classmethod
    def from_dict(cls, d):
        jobs = [QcTask.from_dict(j) for j in d["jobs"]]
        return QcInput(jobs)

    @classmethod
    def from_string(cls, contents):
        qc_contents = contents.split("@@@")
        jobs = [QcTask.from_string(cont) for cont in qc_contents]
        return QcInput(jobs)

    @classmethod
    def from_file(cls, filename):
        with zopen(filename, "rt") as f:
            return cls.from_string(f.read())


class QcOutput(object):

    kcal_per_mol_2_eV = 4.3363E-2

    def __init__(self, filename):
        self.filename = filename
        with zopen(filename, "rt") as f:
            data = f.read()
        chunks = re.split("\n\nRunning Job \d+ of \d+ \S+", data)
        # noinspection PyTypeChecker
        self.data = list(map(self._parse_job, chunks))

    @property
    def final_energy(self):
        return self.data[-1]["energies"][-1][-1]

    @property
    def final_structure(self):
        return self.data[-1]["molecules"][-1]

    @classmethod
    def _expected_successful_pattern(cls, qctask):
        text = ["Convergence criterion met"]
        if "correlation" in qctask.params["rem"]:
            if "ccsd" in qctask.params["rem"]["correlation"]\
                    or "qcisd" in qctask.params["rem"]["correlation"]:
                text.append('CC.*converged')
        if qctask.params["rem"]["jobtype"] == "opt"\
                or qctask.params["rem"]["jobtype"] == "ts":
            text.append("OPTIMIZATION CONVERGED")
        if qctask.params["rem"]["jobtype"] == "freq":
            text.append("VIBRATIONAL ANALYSIS")
        if qctask.params["rem"]["jobtype"] == "gradient":
            text.append("Gradient of SCF Energy")
        return text

    @classmethod
    def _parse_job(cls, output):
        scf_energy_pattern = re.compile("Total energy in the final basis set ="
                                        "\s+(?P<energy>-\d+\.\d+)")
        corr_energy_pattern = re.compile("(?P<name>[A-Z\-\(\)0-9]+)\s+"
                                         "([tT]otal\s+)?[eE]nergy\s+=\s+"
                                         "(?P<energy>-\d+\.\d+)")
        coord_pattern = re.compile("\s*\d+\s+(?P<element>[A-Z][a-zH]*)\s+"
                                   "(?P<x>\-?\d+\.\d+)\s+"
                                   "(?P<y>\-?\d+\.\d+)\s+"
                                   "(?P<z>\-?\d+\.\d+)")
        num_ele_pattern = re.compile("There are\s+(?P<alpha>\d+)\s+alpha "
                                     "and\s+(?P<beta>\d+)\s+beta electrons")
        total_charge_pattern = re.compile("Sum of atomic charges ="
                                          "\s+(?P<charge>\-?\d+\.\d+)")
        scf_iter_pattern = re.compile("\d+\s+(?P<energy>\-\d+\.\d+)\s+"
                                      "(?P<diis_error>\d+\.\d+E[-+]\d+)")
        zpe_pattern = re.compile("Zero point vibrational energy:"
                                 "\s+(?P<zpe>\d+\.\d+)\s+kcal/mol")
        thermal_corr_pattern = re.compile("(?P<name>\S.*\S):\s+"
                                          "(?P<correction>\d+\.\d+)\s+"
                                          "k?cal/mol")
        detailed_charge_pattern = re.compile("(Ground-State )?(?P<method>\w+)( Net)?"
                                             " Atomic Charges")
        nbo_charge_pattern = re.compile("(?P<element>[A-Z][a-z]{0,2})\s*(?P<no>\d+)\s+(?P<charge>\-?\d\.\d+)"
                                        "\s+(?P<core>\-?\d+\.\d+)\s+(?P<valence>\-?\d+\.\d+)"
                                        "\s+(?P<rydberg>\-?\d+\.\d+)\s+(?P<total>\-?\d+\.\d+)"
                                        "(\s+(?P<spin>\-?\d\.\d+))?")
        nbo_wavefunction_type_pattern = re.compile("This is an? (?P<type>\w+\-\w+) NBO calculation")
        bsse_pattern = re.compile("DE, kJ/mol\s+(?P<raw_be>\-?\d+\.?\d+([eE]\d+)?)\s+"
                                  "(?P<corrected_be>\-?\d+\.?\d+([eE]\d+)?)")
        float_pattern = re.compile("\-?\d+\.?\d+([eE]\d+)?$")

        error_defs = (
            (re.compile("Convergence failure"), "Bad SCF convergence"),
            (re.compile("Coordinates do not transform within specified "
                        "threshold"), "autoz error"),
            (re.compile("MAXIMUM OPTIMIZATION CYCLES REACHED"),
                "Geometry optimization failed"),
            (re.compile("\s+[Nn][Aa][Nn]\s+"), "NAN values"),
            (re.compile("energy\s+=\s*(\*)+"), "Numerical disaster"),
            (re.compile("NewFileMan::OpenFile\(\):\s+nopenfiles=\d+\s+"
                        "maxopenfiles=\d+s+errno=\d+"), "Open file error"),
            (re.compile("Application \d+ exit codes: 1[34]\d+"), "Exit Code 134"),
            (re.compile("Negative overlap matrix eigenvalue. Tighten integral "
                        "threshold \(REM_THRESH\)!"), "Negative Eigen"),
            (re.compile("Unable to allocate requested memory in mega_alloc"),
                "Insufficient static memory"),
            (re.compile("Application \d+ exit signals: Killed"),
                "Killed")
        )

        energies = []
        scf_iters = []
        coords = []
        species = []
        molecules = []
        gradients = []
        freqs = []
        vib_freqs = []
        vib_modes = []
        grad_comp = None
        errors = []
        parse_input = False
        parse_coords = False
        parse_scf_iter = False
        parse_gradient = False
        parse_freq = False
        parse_modes = False
        qctask_lines = []
        qctask = None
        jobtype = None
        charge = None
        spin_multiplicity = None
        thermal_corr = dict()
        properly_terminated = False
        pop_method = None
        parse_charge = False
        nbo_available = False
        nbo_charge_header = None
        parse_nbo_charge = False
        charges = dict()
        scf_successful = False
        opt_successful = False
        parse_alpha_homo = False
        parse_alpha_lumo = False
        parse_beta_homo = False
        parse_beta_lumo = False
        current_alpha_homo = None
        current_alpha_lumo = None
        current_beta_homo = None
        homo_lumo = []
        bsse = None
        hiershfiled_pop = False
        for line in output.split("\n"):
            for ep, message in error_defs:
                if ep.search(line):
                    errors.append(message)
            if parse_input:
                if "-" * 50 in line:
                    if len(qctask_lines) == 0:
                        continue
                    else:
                        qctask = QcTask.from_string('\n'.join(qctask_lines))
                        jobtype = qctask.params["rem"]["jobtype"]
                        parse_input = False
                        continue
                qctask_lines.append(line)
            elif parse_coords:
                if "-" * 50 in line:
                    if len(coords) == 0:
                        continue
                    else:
                        if qctask and qctask.ghost_atoms:
                            if isinstance(qctask.mol, Molecule):
                                for i in qctask.ghost_atoms:
                                    species[i] = qctask.mol.sites[i].specie.symbol
                        molecules.append(Molecule(species, coords))
                        coords = []
                        species = []
                        parse_coords = False
                        continue
                if "Atom" in line:
                    continue
                m = coord_pattern.match(line)
                coords.append([float(m.group("x")), float(m.group("y")),
                              float(m.group("z"))])
                species.append(m.group("element"))
            elif parse_scf_iter:
                if "SCF time:  CPU" in line:
                    parse_scf_iter = False
                    continue
                if 'Convergence criterion met' in line:
                    scf_successful = True
                m = scf_iter_pattern.search(line)
                if m:
                    scf_iters[-1].append((float(m.group("energy")),
                                          float(m.group("diis_error"))))
            elif parse_gradient:
                if "Max gradient component" in line:
                    gradients[-1]["max_gradient"] = \
                        float(line.split("=")[1])
                    if grad_comp:
                        if len(grad_comp) == 3:
                            gradients[-1]["gradients"].extend(zip(*grad_comp))
                        else:
                            raise Exception("Gradient section parsing failed")
                    continue
                elif "RMS gradient" in line:
                    gradients[-1]["rms_gradient"] = \
                        float(line.split("=")[1])
                    parse_gradient = False
                    grad_comp = None
                    continue
                elif "." not in line:
                    if grad_comp:
                        if len(grad_comp) == 3:
                            gradients[-1]["gradients"].extend(zip(*grad_comp))
                        else:
                            raise Exception("Gradient section parsing failed")
                    grad_comp = []
                else:
                    grad_line_token = list(line)
                    grad_crowd = False
                    grad_line_final = line
                    for i in range(5, len(line), 12):
                        c = grad_line_token[i]
                        if not c.isspace():
                            grad_crowd = True
                            if ' ' in grad_line_token[i+1: i+6+1] or \
                                    len(grad_line_token[i+1: i+6+1]) < 6:
                                continue
                            grad_line_token[i-1] = ' '
                    if grad_crowd:
                        grad_line_final = ''.join(grad_line_token)
                    grad_comp.append([float(x) for x
                                      in grad_line_final.strip().split()[1:]])
            elif parse_freq:
                if parse_modes:
                    if "TransDip" in line:
                        parse_modes = False
                        for freq, mode in zip(vib_freqs, zip(*vib_modes)):
                            freqs.append({"frequency": freq,
                                          "vib_mode": mode})
                        vib_modes = []
                        continue
                    dis_flat = [float(x) for x in line.strip().split()[1:]]
                    dis_atom = zip(*([iter(dis_flat)]*3))
                    vib_modes.append(dis_atom)
                if "STANDARD THERMODYNAMIC QUANTITIES" in line\
                        or "Imaginary Frequencies" in line:
                    parse_freq = False
                    continue
                if "Frequency:" in line:
                    vib_freqs = [float(vib) for vib
                                 in line.strip().strip().split()[1:]]
                elif "X      Y      Z" in line:
                    parse_modes = True
                    continue
            elif parse_charge:
                if '-'*20 in line:
                    if len(charges[pop_method]) == 0:
                        continue
                    else:
                        pop_method = None
                        parse_charge = False
                else:
                    if len(line.strip()) == 0 or\
                            'Atom' in line:
                        continue
                    else:
                        charges[pop_method].append(float(line.split()[2]))
            elif parse_nbo_charge:
                if '-'*20 in line:
                    if len(charges[pop_method]) == 0:
                        continue
                elif "="*20 in line:
                    pop_method = None
                    parse_nbo_charge = False
                else:
                    m = nbo_charge_pattern.search(line)
                    if m:
                        charges[pop_method].append(float(m.group("charge")))
                    else:
                        raise Exception("Can't find NBO charges")
            elif parse_alpha_homo:
                if "-- Occupied --" in line:
                    continue
                elif "-- Virtual --" in line:
                    parse_alpha_homo = False
                    parse_alpha_lumo = True
                    continue
                else:
                    tokens = line.split()
                    m = float_pattern.search(tokens[-1])
                    if m:
                        current_alpha_homo = float(m.group(0))
                    continue
            elif parse_alpha_lumo:
                current_alpha_lumo = float(line.split()[0])
                parse_alpha_lumo = False
                continue
            elif parse_beta_homo:
                if "-- Occupied --" in line:
                    continue
                elif "-- Virtual --" in line:
                    parse_beta_homo = False
                    parse_beta_lumo = True
                    continue
                else:
                    tokens = line.split()
                    m = float_pattern.search(tokens[-1])
                    if m:
                        current_beta_homo = float(m.group(0))
                    continue
            elif parse_beta_lumo:
                current_beta_lumo = float(line.split()[0])
                parse_beta_lumo = False
                current_homo = max([current_alpha_homo, current_beta_homo])
                current_lumo = min([current_alpha_lumo, current_beta_lumo])
                homo_lumo.append([Energy(current_homo, "Ha").to("eV"),
                                  Energy(current_lumo, "Ha").to("eV")])
                current_alpha_homo = None
                current_alpha_lumo = None
                current_beta_homo = None
                continue
            elif "-" * 50 in line and not (current_alpha_lumo is None):
                homo_lumo.append([Energy(current_alpha_homo, "Ha").to("eV"),
                                  Energy(current_alpha_lumo, "Ha").to("eV")])
                current_alpha_homo = None
                current_alpha_lumo = None
                current_beta_homo = None
                continue
            else:
                if spin_multiplicity is None:
                    m = num_ele_pattern.search(line)
                    if m:
                        spin_multiplicity = int(m.group("alpha")) - \
                            int(m.group("beta")) + 1
                if charge is None:
                    m = total_charge_pattern.search(line)
                    if m:
                        charge = int(float(m.group("charge")))
                if jobtype and jobtype == "freq":
                    m = zpe_pattern.search(line)
                    if m:
                        zpe = float(m.group("zpe"))
                        thermal_corr["ZPE"] = zpe
                    m = thermal_corr_pattern.search(line)
                    if m:
                        thermal_corr[m.group("name")] = \
                            float(m.group("correction"))
                m = bsse_pattern.search(line)
                if m:
                    raw_be = float(m.group("raw_be"))
                    corrected_be = float(m.group("corrected_be"))
                    bsse_fwu = FloatWithUnit(raw_be - corrected_be, "kJ mol^-1")
                    bsse = bsse_fwu.to('eV atom^-1').real
                name = None
                energy = None
                m = scf_energy_pattern.search(line)
                if m:
                    name = "SCF"
                    energy = Energy(m.group("energy"), "Ha").to("eV")
                m = corr_energy_pattern.search(line)
                if m and m.group("name") != "SCF":
                    name = m.group("name")
                    energy = Energy(m.group("energy"), "Ha").to("eV")
                m = detailed_charge_pattern.search(line)
                if m:
                    pop_method = m.group("method").lower()
                    parse_charge = True
                    charges[pop_method] = []
                if nbo_available:
                    if nbo_charge_header is None:
                        m = nbo_wavefunction_type_pattern.search(line)
                        if m:
                            nbo_wavefunction_type = m.group("type")
                            nbo_charge_header_dict = {
                                "closed-shell": "Atom No    Charge        Core      "
                                                "Valence    Rydberg      Total",
                                "open-shell": "Atom No    Charge        Core      "
                                              "Valence    Rydberg      Total      Density"}
                            nbo_charge_header = nbo_charge_header_dict[nbo_wavefunction_type]
                        continue
                    if nbo_charge_header in line:
                        pop_method = "nbo"
                        parse_nbo_charge = True
                        charges[pop_method] = []
                if "N A T U R A L   B O N D   O R B I T A L   A N A L Y S I S" in line:
                    nbo_available = True
                if name and energy:
                    energies.append(tuple([name, energy]))
                if "User input:" in line:
                    parse_input = True
                elif "Standard Nuclear Orientation (Angstroms)" in line:
                    parse_coords = True
                elif "Performing Hirshfeld population analysis" in line:
                    hiershfiled_pop = True
                elif "Hirshfeld: atomic densities completed" in line:
                    hiershfiled_pop = False
                elif ("Cycle       Energy         DIIS Error" in line
                        or "Cycle       Energy        RMS Gradient" in line)\
                        and not hiershfiled_pop:
                    parse_scf_iter = True
                    scf_iters.append([])
                    scf_successful = False
                elif "Gradient of SCF Energy" in line:
                    parse_gradient = True
                    gradients.append({"gradients": []})
                elif "VIBRATIONAL ANALYSIS" in line:
                    parse_freq = True
                elif "Alpha MOs" in line:
                    parse_alpha_homo = True
                    parse_alpha_lumo = False
                elif "Beta MOs" in line:
                    parse_beta_homo = True
                    parse_beta_lumo = False
                elif "Thank you very much for using Q-Chem." in line:
                    properly_terminated = True
                elif "OPTIMIZATION CONVERGED" in line:
                    opt_successful = True
        if charge is None:
            errors.append("Molecular charge is not found")
        elif spin_multiplicity is None:
            errors.append("Molecular spin multipilicity is not found")
        else:
            for mol in molecules:
                if qctask is None or qctask.ghost_atoms is None:
                    mol.set_charge_and_spin(charge, spin_multiplicity)
        for k in thermal_corr.keys():
            v = thermal_corr[k]
            if "Entropy" in k:
                v *= cls.kcal_per_mol_2_eV * 1.0E-3
            else:
                v *= cls.kcal_per_mol_2_eV
            thermal_corr[k] = v

        solvent_method = "NA"
        if qctask:
            if "solvent_method" in qctask.params["rem"]:
                solvent_method = qctask.params["rem"]["solvent_method"]
        else:
            errors.append("No input text")

        if not scf_successful:
            if 'Bad SCF convergence' not in errors:
                errors.append('Bad SCF convergence')

        if jobtype == 'opt':
            if not opt_successful:
                if 'Geometry optimization failed' not in errors:
                    errors.append('Geometry optimization failed')

        if len(errors) == 0:
            for text in cls._expected_successful_pattern(qctask):
                success_pattern = re.compile(text)
                if not success_pattern.search(output):
                    errors.append("Can't find text to indicate success")

        data = {
            "jobtype": jobtype,
            "energies": energies,
            "HOMO/LUMOs": homo_lumo,
            "bsse": bsse,
            'charges': charges,
            "corrections": thermal_corr,
            "molecules": molecules,
            "errors": errors,
            "has_error": len(errors) > 0,
            "frequencies": freqs,
            "gradients": gradients,
            "input": qctask,
            "gracefully_terminated": properly_terminated,
            "scf_iteration_energies": scf_iters,
            "solvent_method": solvent_method
        }
        return data



# coding: utf-8
"""
This module provides objects describing the basic parameters of the 
pseudopotentials used in Abinit, and a parser to instantiate pseudopotential objects..
"""
from __future__ import unicode_literals, division, print_function

import sys
import os
import abc
import collections
import json
import six
import numpy as np

from monty.string import list_strings, is_string
from monty.itertools import iterator_from_slice
from monty.io import FileLock
from monty.collections import AttrDict, Namespace 
from pymatgen.core.periodic_table import PeriodicTable #, Element


__all__ = [
    "Pseudo",
    "PseudoTable",
]

__author__ = "Matteo Giantomassi"
__version__ = "0.1"
__maintainer__ = "Matteo Giantomassi"


_PTABLE = PeriodicTable()

# Tools and helper functions.


def straceback():
    """Returns a string with the traceback."""
    import sys
    import traceback
    return "\n".join((traceback.format_exc(), str(sys.exc_info()[0])))


def _read_nlines(filename, nlines):
    """
    Read at most nlines lines from file filename.
    If nlines is < 0, the entire file is read.
    """
    if nlines < 0:
        with open(filename, 'r') as fh:
            return fh.readlines()

    lines = []
    with open(filename, 'r') as fh:
        for (lineno, line) in enumerate(fh):
            if lineno == nlines: break
            lines.append(line)
        return lines

_l2str = {
    0: "s",
    1: "p",
    2: "d",
    3: "f",
    4: "g",
    5: "h",
    6: "i",
}

_str2l = {v: k for k, v in _l2str.items()}


def l2str(l):
    """Convert the angular momentum l (int) to string."""
    try:
        return _l2str[l]
    except KeyError:
        return "Unknown angular momentum, received l = %s" % l


def str2l(s):
    """Convert a string to the angular momentum l (int)"""
    return _str2l[s]


def read_dojo_report(filename):
    """Helper function to read the DOJO_REPORT from file."""
    with open(filename, "r") as fh:
        lines = fh.readlines()
        try:
            start = lines.index("<DOJO_REPORT>\n")
        except ValueError:
            return {}

        stop = lines.index("</DOJO_REPORT>\n")
        d = json.loads("".join(lines[start+1:stop]))
        return d


class Pseudo(six.with_metaclass(abc.ABCMeta, object)):
    """
    Abstract base class defining the methods that must be 
    implemented by the concrete pseudopotential classes.
    """

    @classmethod
    def as_pseudo(cls, obj):
        """
        Convert obj into a pseudo. Accepts:

            * Pseudo object.
            * string defining a valid path.
        """
        return obj if isinstance(obj, cls) else cls.from_file(obj)

    @staticmethod
    def from_file(filename):
        """
        Return a Pseudo object from filename.
        Note: the parser knows the concrete class that should be instanciated
        """
        return PseudoParser().parse(filename)

    def __repr__(self):
        return "<%s at %s, name = %s>" % (self.__class__.__name__, id(self), self.name)

    def __str__(self):
        """String representation."""
        lines = []
        app = lines.append
        app("<%s: %s>" % (self.__class__.__name__, self.name))
        app("  summary: " + self.summary.strip())
        app("  number of valence electrons: %s" % self.Z_val)
        #FIXME: rewrite the treatment of xc, use XML specs as starting point
        #app("  XC correlation (ixc): %s" % self._pspxc)  #FIXME
        app("  maximum angular momentum: %s" % l2str(self.l_max))
        app("  angular momentum for local part: %s" % l2str(self.l_local))
        if self.isnc:
            app("  radius for non-linear core correction: %s" % self.nlcc_radius)
        app("")

        hint_normal = self.hint_for_accuracy()
        if hint_normal is not None:
            app("  hint for normal accuracy: %s" % str(hint_normal))

        return "\n".join(lines)

    @abc.abstractproperty
    def summary(self):
        """String summarizing the most important properties."""

    @property
    def filepath(self):
        return os.path.abspath(self.path)

    @property
    def name(self):
        """File basename."""
        return os.path.basename(self.filepath)

    @abc.abstractproperty
    def Z(self):
        """The atomic number of the atom."""

    @abc.abstractproperty
    def Z_val(self):
        """Valence charge"""

    @property
    def type(self):
        return self.__class__.__name__

    @property
    def element(self):
        """Pymatgen :class:`Element`."""
        #return Element.from_Z(self.Z)
        try:
            return _PTABLE[self.Z]
        except (KeyError, IndexError):
            return _PTABLE[int(self.Z)]

    @property
    def symbol(self):
        """Element symbol."""
        return self.element.symbol

    @abc.abstractproperty
    def l_max(self):
        """Maximum angular momentum."""

    @abc.abstractproperty
    def l_local(self):
        """Angular momentum used for the local part."""

    @property
    def isnc(self):
        """True if norm-conserving pseudopotential."""
        return isinstance(self, NcPseudo)

    @property
    def ispaw(self):
        """True if PAW pseudopotential."""
        return isinstance(self, PawPseudo)

    #@abc.abstractproperty
    #def xc_type(self):
    #    """XC family e.g LDA, GGA, MGGA."""

    #@abc.abstractproperty
    #def xc_flavor(self):
    #    """XC flavor e.g PW, PW91, PBE."""

    #@property
    #def xc_functional(self):
    #    """XC identifier e.g LDA-PW91, GGA-PBE, GGA-revPBE."""
    #    return "-".join([self.xc_type, self.xc_flavor])

    #@abc.abstractproperty
    #def has_soc(self):
    #    """True if pseudo contains spin-orbit coupling."""

    #@abc.abstractmethod
    #def num_of_projectors(self, l='s'):
    #    """Number of projectors for the angular channel l"""

    #@abc.abstractmethod
    #def generation_mode
    #    """scalar scalar-relativistic, relativistic."""

    def as_dict(self, **kwargs):
        return dict(
            name=self.name,
            type=self.type,
            symbol=self.symbol,
            Z=self.Z,
            Z_val=self.Z_val,
            l_max=self.l_max,
            #nlcc_radius=self.nlcc_radius,
        )

    @property
    def has_dojo_report(self):
        """True if self contains the `DOJO_REPORT` section."""
        return bool(self.dojo_report)

    def delta_factor(self, accuracy="normal"):
        """
        Returns the deltafactor [meV/natom] computed with the given accuracy.
        None if the `Pseudo` does not have info on the deltafactor.
        """
        if not self.has_dojo_report:
            return None
        try:
            return self.dojo_report["delta_factor"][accuracy]["dfact"]
        except KeyError:
            return None

    def read_dojo_report(self):
        """
        Read the `DOJO_REPORT` section and set dojo_report attribute.
        returns {} if section is not present.
        """ 
        self.dojo_report = read_dojo_report(self.path)
        return self.dojo_report

    def write_dojo_report(self, report=None):
        """Write a new `DOJO_REPORT` section to the pseudopotential file."""
        if report is None:
            report = self.dojo_report

        # Create JSON string from report.
        jstring = json.dumps(report, indent=4, sort_keys=True) + "\n"
        #jstring = json.dumps(report, sort_keys=True) + "\n"

        # Read lines from file and insert jstring between the tags.
        with open(self.path, "r") as fh:
            lines = fh.readlines()
            try:
                start = lines.index("<DOJO_REPORT>\n")
            except ValueError:
                start = -1

            if start == -1:
                # DOJO_REPORT was not present.
                lines += ["\n", "<DOJO_REPORT>\n", jstring , "</DOJO_REPORT>\n",]
            else:
                stop = lines.index("</DOJO_REPORT>\n")
                lines.insert(stop, jstring)
                del lines[start+1:stop]

        #  Write new file.
        with FileLock(self.path):
            with open(self.path, "w") as fh:
                fh.writelines(lines)

    def remove_dojo_report(self):
        """Remove the `DOJO_REPORT` section from the pseudopotential file."""
        # Read lines from file and insert jstring between the tags.
        with open(self.path, "r") as fh:
            lines = fh.readlines()
            try:
                start = lines.index("<DOJO_REPORT>\n")
            except ValueError:
                start = -1

            if start == -1:
                return

            stop = lines.index("</DOJO_REPORT>\n")
            if stop == -1:
                return

            del lines[start+1:stop]

        # Write new file.
        with FileLock(self.path):
            with open(self.path, "w") as fh:
                fh.writelines(lines)

    def hint_for_accuracy(self, accuracy="normal"):
        """
        Returns an hint object with parameters such as ecut [Ha] and 
        aug_ratio for given accuracy. Returns None if no hint is available.

        Args:
            accuracy: ["low", "normal", "high"]
        """
        if self.has_dojo_report:
            return Hint.from_dict(self.dojo_report["hints"][accuracy])
        else:
            return None

    @property
    def has_hints(self):
        """True if self provides hints on the cutoff energy."""
        for acc in ["low", "normal", "high"]:
            if self.hint_for_accuracy(acc) is None:
                return False
        return True


class NcPseudo(six.with_metaclass(abc.ABCMeta, object)):
    """
    Abstract class defining the methods that must be implemented
    by the concrete classes representing norm-conserving pseudopotentials.
    """

    @abc.abstractproperty
    def nlcc_radius(self):
        """
        Radius at which the core charge vanish (i.e. cut-off in a.u.).
        Returns 0.0 if nlcc is not used.
        """

    @property
    def has_nlcc(self):
        """True if the pseudo is generated with non-linear core correction."""
        return self.nlcc_radius > 0.0

    @property
    def rcore(self):
        """Radius of the pseudization sphere in a.u."""
        try:
            return self._core
        except AttributeError:
            return None


class PawPseudo(six.with_metaclass(abc.ABCMeta, object)):
    """
    Abstract class that defines the methods that must be implemented
    by the concrete classes representing PAW pseudopotentials.
    """

    #def nlcc_radius(self):
    #    """
    #    Radius at which the core charge vanish (i.e. cut-off in a.u.).
    #    Returns 0.0 if nlcc is not used.
    #    """
    #    return 0.0
    #                                                                           

    #@property
    #def has_nlcc(self):
    #    """True if the pseudo is generated with non-linear core correction."""
    #    return True

    @abc.abstractproperty
    def paw_radius(self):
        """Radius of the PAW sphere in a.u."""

    @property
    def rcore(self):
        """Alias of paw_radius."""
        return self.paw_radius


class AbinitPseudo(Pseudo):
    """
    An AbinitPseudo is a pseudopotential whose file contains an abinit header.
    """
    def __init__(self, path, header):
        """
        Args:
            path: Filename.
            header: :class:`AbinitHeader` instance.
        """
        self.path = path
        self._summary = header.summary

        if hasattr(header, "dojo_report"):
            self.dojo_report = header.dojo_report
        else:
            self.dojo_report = {}

        #self.pspcod  = header.pspcod

        for attr_name, desc in header.items():
            value = header.get(attr_name, None)

            # Hide these attributes since one should always use the public interface.
            setattr(self, "_" + attr_name, value)

    @property
    def summary(self):
        """Summary line reported in the ABINIT header."""
        return self._summary.strip()

    @property
    def Z(self):
        return self._zatom

    @property
    def Z_val(self):
        return self._zion

    @property
    def l_max(self):
        return self._lmax

    @property
    def l_local(self):
        return self._lloc


class NcAbinitPseudo(NcPseudo, AbinitPseudo):
    """Norm-conserving pseudopotential in the Abinit format."""
    @property
    def summary(self):
        return self._summary.strip()

    @property
    def Z(self):
        return self._zatom

    @property
    def Z_val(self):
        """Number of valence electrons."""
        return self._zion

    @property
    def l_max(self):
        return self._lmax

    @property
    def l_local(self):
        return self._lloc

    @property
    def nlcc_radius(self):
        return self._rchrg


class PawAbinitPseudo(PawPseudo, AbinitPseudo):
    """Paw pseudopotential in the Abinit format."""

    @property
    def paw_radius(self):
        return self._r_cut

    #def orbitals(self):


class Hint(collections.namedtuple("Hint", "ecut aug_ratio")):
    """
    Suggested value for the cutoff energy [Hartree units] and the augmentation ratio (PAW pseudo)
    """
    def as_dict(self):
        return {f: getattr(self, f) for f in self._fields}

    @classmethod
    def from_dict(cls, d):
        return cls(**{k: v for k,v in d.items() if not k.startswith("@")})


def _dict_from_lines(lines, key_nums, sep=None):
    """
    Helper function to parse formatted text structured like:

    value1 value2 ... sep key1, key2 ...

    key_nums is a list giving the number of keys for each line. 0 if line should be skipped.
    sep is a string denoting the character that separates the keys from the value (None if
    no separator is present).

    Returns:
        dict{key1 : value1, key2 : value2, ...}

    Raises:
        ValueError if parsing fails.
    """
    if is_string(lines):
        lines = [lines]

    if not isinstance(key_nums, collections.Iterable):
        key_nums = list(key_nums)

    if len(lines) != len(key_nums):
        err_msg = "lines = %s\n key_num =  %s" % (str(lines), str(key_nums))
        raise ValueError(err_msg)

    kwargs = Namespace()

    for (i, nk) in enumerate(key_nums):
        if nk == 0: continue
        line = lines[i]

        tokens = [t.strip() for t in line.split()]
        values, keys = tokens[:nk], "".join(tokens[nk:])
        # Sanitize keys: In some case we might string in for  foo[,bar]
        keys.replace("[", "").replace("]", "")
        keys = keys.split(",")

        if sep is not None:
            check = keys[0][0]
            if check != sep:
                raise ValueError("Expecting separator %s, got %s" % (sep, check))
            keys[0] = keys[0][1:]

        if len(values) != len(keys):
            msg = "line: %s\n len(keys) != len(value)\nkeys: %s\n values:  %s" % (line, keys, values)
            raise ValueError(msg)

        kwargs.update(zip(keys, values))

    return kwargs


class AbinitHeader(dict):
    """Dictionary whose keys can be also accessed as attributes."""
    def __getattr__(self, name):
        try:
            # Default behaviour
            return super(AbinitHeader, self).__getattribute__(name)
        except AttributeError:
            try:
                # Try in the dictionary.
                return self[name]
            except KeyError as exc:
                raise AttributeError(str(exc))


def _int_from_str(string):
    """
    Convert string into integer

    Raise:
        TypeError if string is not a valid integer
    """
    float_num = float(string)
    int_num = int(float_num)
    if float_num == int_num:
        return int_num
    else:
        raise TypeError("Cannot convert string %s to int" % string)


class NcAbinitHeader(AbinitHeader):
    """The abinit header found in the NC pseudopotential files."""
    _attr_desc = collections.namedtuple("att", "default astype")

    _VARS = {
        # Mandatory
        "zatom"        : _attr_desc(None, _int_from_str),
        "zion"         : _attr_desc(None, float),
        "pspdat"       : _attr_desc(None, float),
        "pspcod"       : _attr_desc(None, int),
        "pspxc"        : _attr_desc(None, int),
        "lmax"         : _attr_desc(None, int),
        "lloc"         : _attr_desc(None, int),
        "r2well"       : _attr_desc(None, float),
        "mmax"         : _attr_desc(None, float),
        # Optional variables for non linear-core correction. HGH does not have it.
        "rchrg"        : _attr_desc(0.0,  float),  # radius at which the core charge vanish (i.e. cut-off in a.u.)
        "fchrg"        : _attr_desc(0.0,  float),
        "qchrg"        : _attr_desc(0.0,  float),
    }
    del _attr_desc

    def __init__(self, summary, **kwargs):
        super(NcAbinitHeader, self).__init__()

        # APE uses llocal instead of lloc.
        if "llocal" in kwargs:
            kwargs["lloc"] = kwargs.pop("llocal")

        self.summary = summary.strip()

        for (key, desc) in NcAbinitHeader._VARS.items():
            default, astype = desc.default, desc.astype

            value = kwargs.pop(key, None)

            if value is None:
                value = default
                if default is None:
                    raise RuntimeError("Attribute %s must be specified" % key)
            else:
                try:
                    value = astype(value)
                except:
                    raise RuntimeError("Conversion Error for key %s, value %s" % (key, value))

            self[key] = value

        # Add dojo_report
        self["dojo_report"] = kwargs.pop("dojo_report", {})

        #if kwargs:
        #    raise RuntimeError("kwargs should be empty but got %s" % str(kwargs))

    @staticmethod
    def fhi_header(filename, ppdesc):
        """Parse the FHI abinit header."""
        # Example:
        # Troullier-Martins psp for element  Sc        Thu Oct 27 17:33:22 EDT 1994
        #  21.00000   3.00000    940714                zatom, zion, pspdat
        #    1    1    2    0      2001    .00000      pspcod,pspxc,lmax,lloc,mmax,r2well
        # 1.80626423934776     .22824404341771    1.17378968127746   rchrg,fchrg,qchrg
        lines = _read_nlines(filename, -1)

        try:
            header = _dict_from_lines(lines[:4], [0, 3, 6, 3])
        except ValueError:
            # The last record with rchrg ... seems to be optional.
            header = _dict_from_lines(lines[:3], [0, 3, 6])

        summary = lines[0]

        header["dojo_report"] = read_dojo_report(filename)

        return NcAbinitHeader(summary, **header)

    @staticmethod
    def hgh_header(filename, ppdesc):
        """Parse the HGH abinit header."""
        # Example:
        #Hartwigsen-Goedecker-Hutter psp for Ne,  from PRB58, 3641 (1998)
        #   10   8  010605 zatom,zion,pspdat
        # 3 1   1 0 2001 0  pspcod,pspxc,lmax,lloc,mmax,r2well
        lines = _read_nlines(filename, -1)

        header = _dict_from_lines(lines[:3], [0, 3, 6])
        summary = lines[0]

        header["dojo_report"] = read_dojo_report(filename)

        return NcAbinitHeader(summary, **header)

    @staticmethod
    def gth_header(filename, ppdesc):
        """Parse the GTH abinit header."""
        # Example:
        #Goedecker-Teter-Hutter  Wed May  8 14:27:44 EDT 1996
        #1   1   960508                     zatom,zion,pspdat
        #2   1   0    0    2001    0.       pspcod,pspxc,lmax,lloc,mmax,r2well
        #0.2000000 -4.0663326  0.6778322 0 0     rloc, c1, c2, c3, c4
        #0 0 0                              rs, h1s, h2s
        #0 0                                rp, h1p
        #  1.36 .2   0.6                    rcutoff, rloc
        lines = _read_nlines(filename, -1)

        header = _dict_from_lines(lines[:3], [0, 3, 6])
        summary = lines[0]

        header["dojo_report"] = read_dojo_report(filename)

        return NcAbinitHeader(summary, **header)

    @staticmethod
    def oncvpsp_header(filename, ppdesc):
        """Parse the ONCVPSP abinit header."""
        # Example
        #Li    ONCVPSP  r_core=  2.01  3.02
        #      3.0000      3.0000      140504    zatom,zion,pspd
        #     8     2     1     4   600     0    pspcod,pspxc,lmax,lloc,mmax,r2well
        #  5.99000000  0.00000000  0.00000000    rchrg fchrg qchrg
        #     2     2     0     0     0    nproj
        #     0                 extension_switch
        #   0                        -2.5000025868368D+00 -1.2006906995331D+00
        #     1  0.0000000000000D+00  0.0000000000000D+00  0.0000000000000D+00
        #     2  1.0000000000000D-02  4.4140499497377D-02  1.9909081701712D-02
        lines = _read_nlines(filename, -1)

        header = _dict_from_lines(lines[:3], [0, 3, 6])
        summary = lines[0]

        header.update({'pspdat': header['pspd']})
        header.pop('pspd')

        header["dojo_report"] = read_dojo_report(filename)

        return NcAbinitHeader(summary, **header)

    @staticmethod
    def tm_header(filename, ppdesc):
        """Parse the TM abinit header."""
        # Example:
        #Troullier-Martins psp for element Fm         Thu Oct 27 17:28:39 EDT 1994
        #100.00000  14.00000    940714                zatom, zion, pspdat
        #   1    1    3    0      2001    .00000      pspcod,pspxc,lmax,lloc,mmax,r2well
        #   0   4.085   6.246    0   2.8786493        l,e99.0,e99.9,nproj,rcpsp
        #   .00000000    .0000000000    .0000000000    .00000000   rms,ekb1,ekb2,epsatm
        #   1   3.116   4.632    1   3.4291849        l,e99.0,e99.9,nproj,rcpsp
        #   .00000000    .0000000000    .0000000000    .00000000   rms,ekb1,ekb2,epsatm
        #   2   4.557   6.308    1   2.1865358        l,e99.0,e99.9,nproj,rcpsp
        #   .00000000    .0000000000    .0000000000    .00000000   rms,ekb1,ekb2,epsatm
        #   3  23.251  29.387    1   2.4776730        l,e99.0,e99.9,nproj,rcpsp
        #   .00000000    .0000000000    .0000000000    .00000000   rms,ekb1,ekb2,epsatm
        #   3.62474762267880     .07409391739104    3.07937699839200   rchrg,fchrg,qchrg
        lines = _read_nlines(filename, -1)
        header = []

        for (lineno, line) in enumerate(lines):
            header.append(line)
            if lineno == 2: 
                # Read lmax.
                tokens = line.split()
                pspcod, pspxc, lmax, lloc = map(int, tokens[:4])
                mmax, r2well = map(float, tokens[4:6])
                #if tokens[-1].strip() != "pspcod,pspxc,lmax,lloc,mmax,r2well":
                #    raise RuntimeError("%s: Invalid line\n %s"  % (filename, line))

                lines = lines[3:]
                break

        # TODO
        # Parse the section with the projectors.
        #0   4.085   6.246    0   2.8786493        l,e99.0,e99.9,nproj,rcpsp
        #.00000000    .0000000000    .0000000000    .00000000   rms,ekb1,ekb2,epsatm
        projectors = collections.OrderedDict()
        for idx in range(2*(lmax+1)):
            line = lines[idx]
            if idx % 2 == 0: proj_info = [line,]
            if idx % 2 == 1:
                proj_info.append(line)
                d = _dict_from_lines(proj_info, [5,4])
                projectors[int(d["l"])] = d

        # Add the last line with info on nlcc.
        header.append(lines[idx+1])
        summary = header[0]

        header = _dict_from_lines(header, [0,3,6,3])

        header["dojo_report"] = read_dojo_report(filename)

        return NcAbinitHeader(summary, **header)


class PawAbinitHeader(AbinitHeader):
    """The abinit header found in the PAW pseudopotential files."""
    _attr_desc = collections.namedtuple("att", "default astype")

    _VARS = {
        "zatom"           : _attr_desc(None, _int_from_str),
        "zion"            : _attr_desc(None, float),
        "pspdat"          : _attr_desc(None, float),
        "pspcod"          : _attr_desc(None, int),
        "pspxc"           : _attr_desc(None, int),
        "lmax"            : _attr_desc(None, int),
        "lloc"            : _attr_desc(None, int),
        "mmax"            : _attr_desc(None, int),
        "r2well"          : _attr_desc(None, float),
        "pspfmt"          : _attr_desc(None, str),
        "creatorID"       : _attr_desc(None, int),
        "basis_size"      : _attr_desc(None, int),
        "lmn_size"        : _attr_desc(None, int),
        "orbitals"        : _attr_desc(None, list),
        "number_of_meshes": _attr_desc(None, int),
        "r_cut"           : _attr_desc(None, float), # r_cut(PAW) in the header
        "shape_type"      : _attr_desc(None, int),
        "rshape"          : _attr_desc(None, float),
    }
    del _attr_desc

    def __init__(self, summary, **kwargs):
        super(PawAbinitHeader, self).__init__()

        self.summary = summary.strip()

        for (key, desc) in self._VARS.items():
            default, astype = desc.default, desc.astype

            value = kwargs.pop(key, None)

            if value is None:
                value = default
                if default is None:
                    raise RuntimeError("Attribute %s must be specified" % key)
            else:
                try:
                    value = astype(value)
                except:
                    raise RuntimeError("Conversion Error for key %s, with value %s" % (key, value))

            self[key] = value

        if kwargs:
            raise RuntimeError("kwargs should be empty but got %s" % str(kwargs))

    @staticmethod
    def paw_header(filename, ppdesc):
        """Parse the PAW abinit header."""
        #Paw atomic data for element Ni - Generated by AtomPAW (N. Holzwarth) + AtomPAW2Abinit v3.0.5
        #  28.000  18.000 20061204               : zatom,zion,pspdat
        #  7  7  2 0   350 0.                    : pspcod,pspxc,lmax,lloc,mmax,r2well
        # paw3 1305                              : pspfmt,creatorID
        #  5 13                                  : basis_size,lmn_size
        # 0 0 1 1 2                              : orbitals
        # 3                                      : number_of_meshes
        # 1 3  350 1.1803778368E-05 3.5000000000E-02 : mesh 1, type,size,rad_step[,log_step]
        # 2 1  921 2.500000000000E-03                : mesh 2, type,size,rad_step[,log_step]
        # 3 3  391 1.1803778368E-05 3.5000000000E-02 : mesh 3, type,size,rad_step[,log_step]
        #  2.3000000000                          : r_cut(SPH)
        # 2 0.                

        # Example
        #C  (US d-loc) - PAW data extracted from US-psp (D.Vanderbilt) - generated by USpp2Abinit v2.3.0
        #   6.000   4.000 20090106               : zatom,zion,pspdat
        #  7 11  1 0   560 0.                    : pspcod,pspxc,lmax,lloc,mmax,r2well
        # paw4 2230                              : pspfmt,creatorID
        #  4  8                                  : basis_size,lmn_size
        # 0 0 1 1                                : orbitals
        # 5                                      : number_of_meshes
        # 1 2  560 1.5198032759E-04 1.6666666667E-02 : mesh 1, type,size,rad_step[,log_step]
        # 2 2  556 1.5198032759E-04 1.6666666667E-02 : mesh 2, type,size,rad_step[,log_step]
        # 3 2  576 1.5198032759E-04 1.6666666667E-02 : mesh 3, type,size,rad_step[,log_step]
        # 4 2  666 1.5198032759E-04 1.6666666667E-02 : mesh 4, type,size,rad_step[,log_step]
        # 5 2  673 1.5198032759E-04 1.6666666667E-02 : mesh 5, type,size,rad_step[,log_step]
        #  1.5550009124                          : r_cut(PAW)
        # 3 0.                                   : shape_type,rshape

        #Paw atomic data for element Si - Generated by atompaw v3.0.1.3 & AtomPAW2Abinit v3.3.1
        #  14.000   4.000 20120814               : zatom,zion,pspdat
        #  7      11  1 0   663 0.               : pspcod,pspxc,lmax,lloc,mmax,r2well
        # paw5 1331                              : pspfmt,creatorID
        #  4  8                                  : basis_size,lmn_size
        # 0 0 1 1                                : orbitals
        # 5                                      : number_of_meshes
        # 1 2  663 8.2129718540404674E-04 1.1498160595656655E-02 : mesh 1, type,size,rad_step[,log_step]
        # 2 2  658 8.2129718540404674E-04 1.1498160595656655E-02 : mesh 2, type,size,rad_step[,log_step]
        # 3 2  740 8.2129718540404674E-04 1.1498160595656655E-02 : mesh 3, type,size,rad_step[,log_step]
        # 4 2  819 8.2129718540404674E-04 1.1498160595656655E-02 : mesh 4, type,size,rad_step[,log_step]
        # 5 2  870 8.2129718540404674E-04 1.1498160595656655E-02 : mesh 5, type,size,rad_step[,log_step]
        #  1.5669671236                          : r_cut(PAW)
        # 2 0.                                   : shape_type,rshape
        supported_formats = ["paw3", "paw4", "paw5"]
        if ppdesc.format not in supported_formats:
            raise NotImplementedError("format %s not in %s" % (ppdesc.format, supported_formats))

        lines = _read_nlines(filename, -1)

        summary = lines[0]
        header = _dict_from_lines(lines[:5], [0, 3, 6, 2, 2], sep=":")

        lines = lines[5:]
        # TODO
        # Parse orbitals and number of meshes.
        header["orbitals"] = [int(t) for t in lines[0].split(":")[0].split()]
        header["number_of_meshes"] = num_meshes = int(lines[1].split(":")[0])
        #print filename, header

        # Skip meshes =
        lines = lines[2+num_meshes:]
        #for midx in range(num_meshes):
        #    l = midx + 1

        #print lines[0]
        header["r_cut"] = float(lines[0].split(":")[0])
        #print lines[1]
        header.update(_dict_from_lines(lines[1], [2], sep=":"))

        report = read_dojo_report(filename)
        if report:
            header["dojo_report"] = report

        #print("PAW header\n", header)
        return PawAbinitHeader(summary, **header)


class PseudoParserError(Exception):
    """Base Error class for the exceptions raised by `PseudoParser`"""


class PseudoParser(object):
    """
    Responsible for parsing pseudopotential files and returning pseudopotential objects.

    Usage::
        pseudo = PseudoParser().parse("filename")
    """
    Error = PseudoParserError

    # Supported values of pspcod
    ppdesc = collections.namedtuple("ppdesc", "pspcod name psp_type format")

    # TODO Recheck
    _PSPCODES = collections.OrderedDict( {
        1: ppdesc(1, "TM",  "NC", None),
        2: ppdesc(2, "GTH",  "NC", None),
        3: ppdesc(3, "HGH", "NC", None),
        #4: ppdesc(4, "NC",     , None),
        #5: ppdesc(5, "NC",     , None),
        6: ppdesc(6, "FHI", "NC", None),
        7: ppdesc(6, "PAW_abinit_text", "PAW", None),
        8: ppdesc(8, "ONCVPSP", "NC", None),
       10: ppdesc(10, "HGHK", "NC", None),
    })
    del ppdesc
    # renumber functionals from oncvpsp todo confrim that 3 is 2
    _FUNCTIONALS = {1: {'n': 4, 'name': 'Wigner'},
                    2: {'n': 5, 'name': 'HL'},
                    3: {'n': 2, 'name': 'PWCA'},
                    4: {'n': 11, 'name': 'PBE'}}

    def __init__(self):
        # List of files that have been parsed succesfully.
        self._parsed_paths = []

        # List of files that could not been parsed.
        self._wrong_paths  = []

    def scan_directory(self, dirname, exclude_exts=(), exclude_fnames=()):
        """
        Analyze the files contained in directory dirname.

        Args:
            dirname: directory path
            exclude_exts: list of file extensions that should be skipped.
            exclude_fnames: list of file names that should be skipped.

        Returns:
            List of pseudopotential objects.
        """
        for (i, ext) in enumerate(exclude_exts):
            if not ext.strip().startswith("."):
                exclude_exts[i] =  "." + ext.strip()

        # Exclude files depending on the extension.
        paths = []
        for fname in os.listdir(dirname):
            root, ext = os.path.splitext(fname)
            path = os.path.join(dirname, fname)
            if (ext in exclude_exts or fname in exclude_fnames or 
                fname.startswith(".") or not os.path.isfile(path)): continue
            paths.append(path)

        pseudos = []
        for path in paths:
            # Parse the file and generate the pseudo.
            try:
                pseudo = self.parse(path)
            except:
                pseudo = None

            if pseudo is not None:
                pseudos.append(pseudo)
                self._parsed_paths.extend(path)
            else:
                self._wrong_paths.extend(path)

        return pseudos

    def read_ppdesc(self, filename):
        """
        Read the pseudopotential descriptor from file filename.

        Returns:
            Pseudopotential descriptor. None if filename is not a valid pseudopotential file.

        Raises:
            `PseudoParserError` if fileformat is not supported.
        """
        if filename.endswith(".xml"):
            raise self.Error("XML pseudo not supported yet")

        else:
            # Assume file with the abinit header.
            lines = _read_nlines(filename, 80)

            for (lineno, line) in enumerate(lines):

                if lineno == 2:
                    try:
                        tokens = line.split()
                        pspcod, pspxc = map(int, tokens[:2])
                    except:
                        msg = "%s: Cannot parse pspcod, pspxc in line\n %s" % (filename, line)
                        sys.stderr.write(msg)
                        return None

                    #if tokens[-1].strip().replace(" ","") not in ["pspcod,pspxc,lmax,lloc,mmax,r2well",
                    #                              "pspcod,pspxc,lmax,llocal,mmax,r2well"]:
                    #    raise self.Error("%s: Invalid line\n %s"  % (filename, line))
                    #    return None

                    if pspcod not in self._PSPCODES:
                        raise self.Error("%s: Don't know how to handle pspcod %s\n" % (filename, pspcod))

                    ppdesc = self._PSPCODES[pspcod]

                    if pspcod == 7:
                        # PAW -> need to know the format pspfmt
                        tokens = lines[lineno+1].split()
                        pspfmt, creatorID = tokens[:2]
                        #if tokens[-1].strip() != "pspfmt,creatorID":
                        #    raise self.Error("%s: Invalid line\n %s" % (filename, line))
                        #    return None

                        ppdesc = ppdesc._replace(format = pspfmt)

                    return ppdesc

            return None

    def parse(self, filename):
        """
        Read and parse a pseudopotential file. Main entry point for client code.

        Returns: 
            pseudopotential object or None if filename is not a valid pseudopotential file.
        """
        path = os.path.abspath(filename)

        # Only PAW supports XML at present.
        if filename.endswith(".xml"):
            return PawXmlSetup(path)

        ppdesc = self.read_ppdesc(path)

        if ppdesc is None: 
            return None

        psp_type = ppdesc.psp_type

        parsers = {
            "FHI"            : NcAbinitHeader.fhi_header,
            "GTH"            : NcAbinitHeader.gth_header,
            "TM"             : NcAbinitHeader.tm_header,
            "HGH"            : NcAbinitHeader.hgh_header,
            "HGHK"           : NcAbinitHeader.hgh_header,
            "ONCVPSP"        : NcAbinitHeader.oncvpsp_header,
            "PAW_abinit_text": PawAbinitHeader.paw_header,
        }

        try:
            header = parsers[ppdesc.name](path, ppdesc)
        except Exception as exc:
            raise self.Error(path + ":\n" + straceback())

        root, ext = os.path.splitext(path)

        if psp_type == "NC":
            pseudo = NcAbinitPseudo(path, header)
        elif psp_type == "PAW":
            pseudo = PawAbinitPseudo(path, header)
        else:
            raise NotImplementedError("psp_type not in [NC, PAW]")

        return pseudo


#TODO use RadialFunction from pseudo_dojo.
class RadialFunction(collections.namedtuple("RadialFunction", "mesh values")):
    pass


class PawXmlSetup(Pseudo, PawPseudo):
    def __init__(self, filepath):
        # FIXME
        self.dojo_report = {}
        self.path = os.path.abspath(filepath)

        # Get the XML root (this trick is used to that the object is pickleable).
        root = self.root

        # Get the version of the XML format
        self.paw_setup_version = root.get("version")

        # Info on the atom.
        atom_attrib = root.find("atom").attrib

        #self._symbol = atom_attrib["symbol"]
        self._zatom = int(float(atom_attrib["Z"]))
        self.core, self.valence = map(float, [atom_attrib["core"], atom_attrib["valence"]])

        #xc_info = root.find("atom").attrib
        #self.xc_type, self.xc_name  = xc_info["type"], xc_info["name"]
        #self.ae_energy = {k: float(v) for k,v in root.find("ae_energy").attrib.items()}

        # Old XML files do not define this field!
        # In this case we set the PAW radius to None.
        #self._paw_radius = float(root.find("PAW_radius").attrib["rpaw"])

        pawr_element = root.find("PAW_radius")
        self._paw_radius = None
        if pawr_element is not None:
            self._paw_radius = float(pawr_element.attrib["rpaw"])

        #<valence_states>
        #  <state n="2" l="0" f="2"  rc="1.10" e="-0.6766" id="N-2s"/>
        #  <state n="2" l="1" f="3"  rc="1.10" e="-0.2660" id="N-2p"/>
        #  <state       l="0"        rc="1.10" e=" 0.3234" id="N-s1"/>
        #  <state       l="1"        rc="1.10" e=" 0.7340" id="N-p1"/>
        #  <state       l="2"        rc="1.10" e=" 0.0000" id="N-d1"/>
        #</valence_states>
        #
        # The valence_states element contains several state elements.
        # For this setup, the first two lines describe bound eigenstates
        # with occupation numbers and principal quantum numbers.
        # Notice, that the three additional unbound states should have no f and n attributes.
        # In this way, we know that only the first two bound states (with f and n attributes)
        # should be used for constructing an initial guess for the wave functions.

        self.valence_states = {}
        for node in root.find("valence_states"):
            attrib = AttrDict(node.attrib)
            assert attrib.id not in self.valence_states
            self.valence_states[attrib.id] = attrib
        #print(self.valence_states)

        # Parse the radial grids
        self.rad_grids = {}
        for node in root.findall("radial_grid"):
            grid_params = node.attrib
            gid = grid_params["id"]
            assert gid not in self.rad_grids

            self.rad_grids[id] = self._eval_grid(grid_params)

    def __getstate__(self):
        """
        Return state is pickled as the contents for the instance.
                                                                                      
        In this case we just remove the XML root element process since Element object cannot be pickled.
        """
        return {k: v for k, v in self.__dict__.items() if k not in ["_root"]}

    @property
    def root(self):
        try:
            return self._root
        except AttributeError:
            from xml.etree import cElementTree as Et
            tree = Et.parse(self.filepath)
            self._root = tree.getroot()
            return self._root

    @property
    def Z(self):
        return self._zatom

    @property
    def Z_val(self):
        """Number of valence electrons."""
        return self.valence

    # FIXME
    @property
    def l_max(self):
        """Maximum angular momentum."""
        return None
                                                        
    @property
    def l_local(self):
        """Angular momentum used for the local part."""
        return None

    @property
    def summary(self):
        """String summarizing the most important properties."""
        return ""

    @property
    def paw_radius(self):
        return self._paw_radius

    @staticmethod
    def _eval_grid(grid_params):
        """
        This function receives a dictionary with the parameters defining the
        radial mesh and returns a `ndarray` with the mesh
        """
        eq = grid_params.get("eq").replace(" ", "")
        istart, iend = int(grid_params.get("istart")), int(grid_params.get("iend"))
        indices = list(range(istart, iend+1))

        if eq == 'r=a*exp(d*i)':
            a, d = float(grid_params['a']), float(grid_params['d'])
            mesh = [a * np.exp(d * i) for i in indices]

        elif eq == 'r=a*i/(n-i)':
            a, n = float(grid_params['a']), float(grid_params['n'])
            mesh = [a * i / (n - i) for i in indices]

        elif eq == 'r=a*(exp(d*i)-1)':
            a, d = float(grid_params['a']), float(grid_params['d'])
            mesh = [a * (np.exp(d * i) - 1.0) for i in indices]

        elif eq == 'r=d*i':
            d = float(grid_params['d'])
            mesh = [d * i for i in indices]

        elif eq == 'r=(i/n+a)^5/a-a^4':
            a, n = float(grid_params['a']), float(grid_params['n'])
            mesh = [(i / n + a)**5 / a - a**4 for i in indices]

        else:
            raise ValueError('Unknown grid type: %s' % eq)

        return np.array(mesh)

    def _parse_radfunc(self, func_name):
        """Parse the first occurence of func_name in the XML file."""
        node = self.root.find(func_name)
        grid = node.attrib["grid"]
        values = np.array([float(s) for s in node.text.split()])

        return self.rad_grids[grid], values, node.attrib

    def _parse_all_radfuncs(self, func_name):
        """Parse all the nodes with tag func_name in the XML file."""
        for node in self.root.findall(func_name):
            grid = node.attrib["grid"]
            values = np.array([float(s) for s in node.text.split()])

            yield self.rad_grids[grid], values, node.attrib

    @property
    def ae_core_density(self):
        """The all-electron radial density."""
        try:
            return self._ae_core_density

        except AttributeError:
            mesh, values, attrib = self._parse_radfunc("ae_core_density")
            self._ae_core_density = RadialFunction(mesh, values)
            return self._ae_core_density

    @property
    def pseudo_core_density(self):
        """The pseudized radial density."""
        try:
            return self._pseudo_core_density

        except AttributeError:
            mesh, values, attrib = self._parse_radfunc("pseudo_core_density")
            self._pseudo_core_density = RadialFunction(mesh, values)
            return self._pseudo_core_density

    @property
    def ae_partial_waves(self):
        """Dictionary with the AE partial waves indexed by state."""
        try:
            return self._ae_partial_waves

        except AttributeError:
            self._ae_partial_waves = {}
            for (mesh, values, attrib) in self._parse_all_radfuncs("ae_partial_wave"):
                state = attrib["state"]
                val_state = self.valence_states[state]
                self._ae_partial_waves[state] = RadialFunction(mesh, values)
                #print("val_state", val_state)

            return self._ae_partial_waves

    @property
    def pseudo_partial_waves(self):
        """Dictionary with the pseudo partial waves indexed by state."""
        try:
            return self._pseudo_partial_waves

        except AttributeError:
            self._pseudo_partial_waves = {}
            for (mesh, values, attrib) in self._parse_all_radfuncs("pseudo_partial_wave"):
                state = attrib["state"]
                val_state = self.valence_states[state]
                self._pseudo_partial_waves[state] = RadialFunction(mesh, values)

            return self._pseudo_partial_waves

    @property
    def projector_functions(self):
        """Dictionary with the PAW projectors indexed by state."""
        try:
            return self._projector_functions

        except AttributeError:
            self._projector_functions = {}
            for (mesh, values, attrib) in self._parse_all_radfuncs("projector_function"):
                state = attrib["state"]
                val_state = self.valence_states[state]
                self._projector_functions[state] = RadialFunction(mesh, values)

            return self._projector_functions

    def plot_densities(self, **kwargs):
        """
        Plot the PAW densities.

        ================  ==============================================================
        kwargs            Meaning
        ================  ==============================================================
        title             Title of the plot (Default: "Densities").
        show              True to show the figure (Default).
        savefig           'abc.png' or 'abc.eps' to save the figure to a file.
        ================  ==============================================================

        Returns:
            `matplotlib` figure
        """
        title = kwargs.pop("title", "Densities")
        show = kwargs.pop("show", True)
        savefig = kwargs.pop("savefig", None)

        import matplotlib.pyplot as plt

        fig = plt.figure()

        ax = fig.add_subplot(1,1,1)
        ax.grid(True)
        ax.set_xlabel('r [Bohr]')
        #ax.set_ylabel('density')

        for i, den_name in enumerate(["ae_core_density", "pseudo_core_density"]):
            rden = getattr(self, den_name)
            label = "$n_c$" if i == 1 else "$\\tilde{n}_c$"
            ax.plot(rden.mesh, rden.mesh * rden.values, label=label, lw=2)

        plt.legend(loc="best")

        if title is not None:
            fig.suptitle(title)

        if show:
            plt.show()

        if savefig:
            fig.savefig(savefig)

        return fig

    def plot_waves(self, **kwargs):
        """
        Plot the AE and the pseudo partial waves.

        ================  ==============================================================
        kwargs            Meaning
        ================  ==============================================================
        title             Title of the plot (Default: "Partial Waves").
        show              True to show the figure (Default).
        savefig           'abc.png' or 'abc.eps' to save the figure to a file.
        ================  ==============================================================

        Returns:
            `matplotlib` figure
        """
        title = kwargs.pop("title", "Partial Waves")
        show = kwargs.pop("show", True)
        savefig = kwargs.pop("savefig", None)

        import matplotlib.pyplot as plt

        fig = plt.figure()

        ax = fig.add_subplot(1,1,1)
        ax.grid(True)
        ax.set_xlabel("r [Bohr]")
        ax.set_ylabel("$r\phi,\\, r\\tilde\phi\, [Bohr]^{-\\frac{1}{2}}$")

        ax.axvline(x=self.paw_radius, linewidth=2, color='k', linestyle="--")
        #ax.annotate("$r_c$", xy=(self.paw_radius + 0.1, 0.1))

        for state, rfunc in self.pseudo_partial_waves.items():
            ax.plot(rfunc.mesh, rfunc.mesh * rfunc.values, lw=2, label="PS-WAVE: " + state)

        for state, rfunc in self.ae_partial_waves.items():
            ax.plot(rfunc.mesh, rfunc.mesh * rfunc.values, lw=2, label="AE-WAVE: " + state)

        plt.legend(loc="best")

        if title is not None:
            fig.suptitle(title)

        if show:
            plt.show()

        if savefig:
            fig.savefig(savefig)

        return fig

    def plot_projectors(self, **kwargs):
        """
        Plot the PAW projectors.

        ================  ==============================================================
        kwargs            Meaning
        ================  ==============================================================
        title             Title of the plot (Default: "Projectors").
        show              True to show the figure (Default).
        savefig           'abc.png' or 'abc.eps' to save the figure to a file.
        ================  ==============================================================

        Returns:
            `matplotlib` figure
        """
        title = kwargs.pop("title", "Projectors")
        show = kwargs.pop("show", True)
        savefig = kwargs.pop("savefig", None)

        import matplotlib.pyplot as plt

        fig = plt.figure()

        ax = fig.add_subplot(1,1,1)
        ax.grid(True)
        ax.set_xlabel('r [Bohr]')
        ax.set_ylabel("$r\\tilde p\, [Bohr]^{-\\frac{1}{2}}$")

        ax.axvline(x=self.paw_radius, linewidth=2, color='k', linestyle="--")
        #ax.annotate("$r_c$", xy=(self.paw_radius + 0.1, 0.1))

        for state, rfunc in self.projector_functions.items():
            ax.plot(rfunc.mesh, rfunc.mesh * rfunc.values, label="TPROJ: " + state)

        plt.legend(loc="best")

        if title is not None:
            fig.suptitle(title)

        if show:
            plt.show()

        if savefig:
            fig.savefig(savefig)

        return fig

    #def plot_potentials(self, **kwargs):
    #    """
    #        ================  ==============================================================
    #        kwargs            Meaning
    #        ================  ==============================================================
    #        title             Title of the plot (Default: None).
    #        show              True to show the figure (Default).
    #        savefig           'abc.png' or 'abc.eps' to save the figure to a file.
    #        ================  ==============================================================

    #    Returns:
    #        `matplotlib` figure
    #    """
    #    title = kwargs.pop("title", "Potentials")
    #    show = kwargs.pop("show", True)
    #    savefig = kwargs.pop("savefig", None)

    #    import matplotlib.pyplot as plt

    #    fig = plt.figure()

    #    ax = fig.add_subplot(1,1,1)
    #    ax.grid(True)
    #    ax.set_xlabel('r [Bohr]')
    #    ax.set_ylabel('density')
    #    ax.axvline(x=self.paw_radius, linewidth=2, color='k', linestyle="--")
    #    ax.annotate("$r_c$", xy=(self.paw_radius + 0.1, 0.1))

    #    for state, rfunc in self.potentials.items():
    #        ax.plot(rfunc.mesh, rfunc.values, label="TPROJ: " + state)

    #    plt.legend(loc="best")

    #    if title is not None:
    #        fig.suptitle(title)

    #    if show:
    #        plt.show()

    #    if savefig:
    #        fig.savefig(savefig)

    #    return fig


class PseudoTable(collections.Sequence):
    """
    Define the pseudopotentials from the element table.
    Individidual elements are accessed by name, symbol or atomic number.

    For example, the following all retrieve iron:

    print elements[26]
    Fe
    print elements.Fe
    Fe
    print elements.symbol('Fe')
    Fe
    print elements.name('iron')
    Fe
    print elements.isotope('Fe')
    Fe
    """
    @classmethod
    def as_table(cls, items):
        """
        Return an instance of `PseudoTable` from the iterable items.
        """ 
        if isinstance(items, cls): return items
        return cls(items)

    def __init__(self, pseudos):
        """
        Args:
            pseudos:
                List of pseudopotentials or filepaths
        """
        # Store pseudos in a default dictionary with z as key.
        # Note that we can have more than one pseudo for given z.
        # hence the values are lists of pseudos.
        if not isinstance(pseudos, collections.Iterable):
            pseudos = [pseudos]

        if is_string(pseudos[0]):
            pseudos = list_strings(pseudos)

        self._pseudos_with_z = collections.defaultdict(list)

        for pseudo in pseudos:
            p = pseudo
            if not isinstance(pseudo, Pseudo):
                p = Pseudo.from_file(pseudo)

            self._pseudos_with_z[p.Z].append(p)

        for z in self.zlist:
            pseudo_list = self._pseudos_with_z[z]
            symbols = [p.symbol for p in pseudo_list]
            symbol = symbols[0]
            if any(symb != symbol for symb in symbols):
                raise ValueError("All symbols must be equal while they are: %s" % str(symbols))

            setattr(self, symbol, pseudo_list)

    def __getitem__(self, Z):
        """
        Retrieve pseudos for the atomic number z.
        Accepts both int and slice objects.
        """
        if isinstance(Z, slice):
            assert Z.stop is not None
            pseudos = []
            for znum in iterator_from_slice(Z):
                pseudos.extend(self._pseudos_with_z[znum])
            return pseudos
        else:
            return self._pseudos_with_z[Z]

    def __len__(self):
        return len(list(self.__iter__()))

    def __iter__(self):
        """Process the elements in Z order."""
        for z in self.zlist:
            for pseudo in self._pseudos_with_z[z]:
                yield pseudo

    def __repr__(self):
        return "<%s at %s>" % (self.__class__.__name__, id(self))

    def __str__(self):
        lines = []
        app = lines.append
        app("<%s, len=%d>" % (self.__class__.__name__, len(self)))

        for pseudo in self:
            app(str(pseudo))

        return "\n".join(lines)

    @property
    def allnc(self):
        """True if all pseudos are norm-conserving."""
        return all(p.isnc for p in self)

    @property
    def allpaw(self):
        """True if all pseudos are PAW."""
        return all(p.ispaw for p in self)

    @property
    def zlist(self):
        """Ordered list with the atomic numbers available in the table."""
        zlist = list(self._pseudos_with_z.keys())
        zlist.sort()
        return zlist

    def as_dict(self, **kwargs):
        d = {}
        for p in self:
            k, count = p.name, 1
            # Handle multiple-pseudos with the same name!
            while k not in d:
                k += k + "#" + str(count)
                count += 1
            d.update({k, p.as_dict()})
        return d

    def is_complete(self, zmax=118):
        """
        True if table is complete i.e. all elements with Z < zmax
        have at least on pseudopotential
        """
        for z in range(1, zmax):
            if not self[z]: return False
        return True

    def pseudos_with_symbol(self, symbol):
        """
        Return the list of pseudopotentials in the table the with given symbol.
        Return an empty list if no pseudo is avaiable
        """
        try:
            return getattr(self, str(symbol))
        except AttributeError:
            return []

    def pseudo_from_name(self, name):
        """Return the pseudo in the table with the given name"""
        for pseudo in self:
            if pseudo.name == name:
                return pseudo
        return None

    def list_properties(self, *props, **kw):
        """
        Print a list of elements with the given set of properties.

        Args:
            *prop1*, *prop2*, ... : string
                Name of the properties to print
            *format*: string
                Template for displaying the element properties, with one
                % for each property.

        For example, print a table of mass and density.

        from periodictable import elements
        elements.list_properties('symbol','mass','density', format="%-2s: %6.2f u %5.2f g/cm^3")
        H :   1.01 u   0.07 g/cm^3
        He:   4.00 u   0.12 g/cm^3
        Li:   6.94 u   0.53 g/cm^3
        ...
        Bk: 247.00 u  14.00 g/cm^3
        """
        format = kw.pop('format', None)
        assert len(kw) == 0

        for pseudo in self:
            try:
                values = tuple(getattr(pseudo, p) for p in props)
            except AttributeError:
                # Skip elements which don't define all the attributes
                continue

            # Skip elements with a value of None
            if any(v is None for v in values):
                continue

            if format is None:
                print(" ".join(str(p) for p in values))
            else:
                try:
                    print(format % values)
                except:
                    print("format",format,"args",values)
                    raise

    #def print_table(self, stream=sys.stdout, filter_function=None):
    #    """
    #    A pretty ASCII printer for the periodic table, based on some filter_function.

    #    Args:
    #        filter_function:
    #            A filtering function that take a Pseudo as input and returns a boolean.
    #            For example, setting filter_function = lambda el: el.Z_val > 2 will print
    #            a periodic table containing only pseudos with Z_val > 2.
    #    """
    #    for row in range(1, 10):
    #        rowstr = []
    #        for group in range(1, 19):
    #            el = Element.from_row_and_group(row, group)
    #            if el and ((not filter_function) or filter_function(el)):
    #                rowstr.append("{:3s}".format(el.symbol))
    #            else:
    #                rowstr.append("   ")
    #        print(" ".join(rowstr))

    def sorted(self, attrname, reverse=False):
        """Sort the table according to the value of attribute attrname."""
        attrs = []
        for i, pseudo in self:
            try:
                a = getattr(pseudo, attrname)
            except AttributeError:
                a = np.inf
            attrs.append((i, a))

        # Sort attrs, and build new table with sorted pseudos.
        attrs = sorted(attrs, key=lambda t: t[1], reverse=reverse)
        return PseudoTable([self[a[0]] for a in attrs])

    def select(self, condition):
        """
        Select only those pseudopotentials for which condition is True.

        Args:
            condition:
                Function that accepts a `Pseudo` object and returns True or False.
        """
        return PseudoTable([p for p in self if condition(p)])

    def with_dojo_report(self):
        """Select pseudos containing the DOJO_REPORT section."""
        return self.select(condition=lambda p: p.has_dojo_report)




# coding: utf-8

from __future__ import division, unicode_literals

"""
This module defines the FeffInputSet abstract base class and a concrete
implementation for the Materials Project.  The basic concept behind an input
set is to specify a scheme to generate a consistent set of Feff inputs from a
structure without further user intervention. This ensures comparability across
runs.
"""

import six

__author__ = "Alan Dozier"
__credits__ = "Anubhav Jain, Shyue Ping Ong"
__copyright__ = "Copyright 2011, The Materials Project"
__version__ = "1.0.3"
__maintainer__ = "Alan Dozier"
__email__ = "adozier@uky.edu"
__date__ = "April 7, 2013"

import os
import abc

from monty.serialization import loadfn

from pymatgen.io.feffio import FeffAtoms, FeffTags, FeffPot, Header


class AbstractFeffInputSet(six.with_metaclass(abc.ABCMeta, object)):
    """
    Abstract base class representing a set of Feff input parameters.
    The idea is that using a FeffInputSet, a complete set of input files
    (feffPOT,feffXANES, feffEXAFS, ATOMS, feff.inp)set_
    can be generated in an automated fashion for any structure.
    """

    @abc.abstractmethod
    def get_feff_atoms(self, structure, central_atom):
        """
        Returns Atoms string from a structure that goes in feff.inp file.

        Args:
            structure: pymatgen structure object
            central_atom: atom symbol string for absorbing atom

        Returns:
            FeffAtoms object.
        """
        return

    @abc.abstractmethod
    def get_feff_tags(self, calc_type):
        """
        Returns standard calculation paramters for either an FEFF XANES or
        EXAFS input.

        Args:
            calc_type: At this time either 'XANES' or 'EXAFS' string is
                supported for K shell excitation. In the future this will be
                expanded to include other shells and material class
                differentiation.
        """
        return

    @abc.abstractmethod
    def get_feff_pot(self, structure, central_atom):
        """
        Returns POTENTIAL section used in feff.inp from a structure.

        Args:
            structure: pymatgen structure object
            central_atom: atom symbol string for absorbing atom
        """
        return

    @abc.abstractmethod
    def get_header(self, structure, source, comment):
        """
        Returns header to be used in feff.inp file from a pymatgen structure

        Args:
            structure: A pymatgen structure object
            source: Source identifier used to create structure, can be defined
                however user wants to organize structures, calculations, etc.
                example would be Materials Project material ID number.
        """
        return

    def get_all_feff_input(self, structure, calc_type, source, central_atom,
                           comment=''):
        """
        Returns all input files as a dict of {filename: feffio object}

        Args:
            structure: Structure object
            calc_type: At this time either 'XANES' or 'EXAFS' string is
                supported for K shell excitation. In the future this will be
                expanded to inlude other shells and material class
                differentiation.
            source: Source identifier used to create structure, can be defined
                however user wants to organize structures, calculations, etc.
                example would be Materials Project material ID number.
            central_atom: Atom symbol string for absorbing atom
            comment: Comment to appear in Header.

        Returns:
            dict of objects used to create feff.inp file i.e. Header, FeffTags,
            FeffPot, FeffAtoms
        """

        feff = {"HEADER": self.get_header(structure, source, comment),
                "PARAMETERS": self.get_feff_tags(calc_type),
                "POTENTIALS": self.get_feff_pot(structure, central_atom),
                "ATOMS": self.get_feff_atoms(structure, central_atom)}

        return feff

    def write_input(self, structure, calc_type, source, central_atom,
                    comment='', output_dir=".", make_dir_if_not_present=True):
        """
        Writes a set of FEFF input to a directory.

        Args:
            structure: Structure object
            calc_type: At this time either 'XANES' or 'EXAFS' string is
                supported for K shell excitation. In the future this will be
                expanded to include other shells and material class
                differentiation.
            source: Source identifier used to create structure, can be defined
                however user wants to organize structures, calculations, etc.
                example would be Materials Project material ID number.
            central_atom: Atom symbol string for absorbing atom
            output_dir: Directory to output the FEFF input files
            comment: comment for Header
            make_dir_if_not_present: Set to True if you want the directory (
                and the whole path) to be created if it is not present.
        """

        if make_dir_if_not_present and not os.path.exists(output_dir):
            os.makedirs(output_dir)

        feff = self.get_all_feff_input(structure, calc_type, source,
                                       central_atom, comment)

        feff_input = "\n\n".join(str(feff[f]) for f in ["HEADER", "PARAMETERS",
                                 "POTENTIALS", "ATOMS"])

        for k, v in six.iteritems(feff):
            with open(os.path.join(output_dir, k), "w") as f:
                f.write(str(v))

        with open(os.path.join(output_dir, "feff.inp"), "w") as f:
            f.write(feff_input)
        f.close()

    def as_dict(self, structure, calc_type, source, central_atom,
                comment=''):
        """Creates a feff.inp dictionary as a string"""

        feff = self.get_all_feff_input(structure, calc_type, source,
                                       central_atom, comment)
        feff_input = "\n\n".join(str(feff[f]) for f in ["HEADER", "PARAMETERS",
                                 "POTENTIALS", "ATOMS"])
        return {'@module': self.__class__.__module__,
                '@class': self.__class__.__name__,
                'feff.inp': feff_input}

    @staticmethod
    def from_dict(d):
        """Return feff.inp from a dictionary string representation"""
        return d['feff.inp']


class FeffInputSet(AbstractFeffInputSet):
    """
    Standard implementation of FeffInputSet, which can be extended by specific
    implementations.

    Args:
        name: The name of a grouping of input parameter sets such as
            "MaterialsProject".
    """

    def __init__(self, name):
        self.name = name
        module_dir = os.path.dirname(os.path.abspath(__file__))
        config = loadfn(os.path.join(module_dir, "FeffInputSets.yaml"))
        self.xanes_settings = config[self.name + "feffXANES"]
        self.exafs_settings = config[self.name + "feffEXAFS"]

    def get_header(self, structure, source='', comment=''):
        """
        Creates header string from structure object

        Args:
            structure: A pymatgen structure object
            source: Source identifier used to create structure, can be defined
                however user wants to organize structures, calculations, etc.
                example would be Materials Project material ID number.
            comment: comment to include in header

        Returns:
            Header object to be used in feff.inp file from a pymatgen structure
        """
        return Header(structure, source, comment)

    def get_feff_tags(self, calc_type):
        """
        Reads standard parameters for XANES or EXAFS calculation
        from FeffInputSets.yaml file.

        Args:
            calc_type: At this time either 'XANES' or 'EXAFS' string is
                supported for K shell excitation. In the future this will be
                expanded to include other shells and material class
                differentiation.

        Returns:
            FeffTags object
        """

        if calc_type.upper() == "XANES":
            fefftags = FeffTags(self.xanes_settings)
        elif calc_type.upper() == "EXAFS":
            fefftags = FeffTags(self.exafs_settings)
        else:
            raise ValueError("{} is not a valid calculation type"
                             .format(calc_type))

        return fefftags

    def get_feff_pot(self, structure, central_atom):
        """
        Creates string representation of potentials used in POTENTIAL file and
        feff.inp.

        Args:
            structure: pymatgen structure object
            central_atom: atom symbol string for absorbing atom

        Returns:
            FeffPot object
        """
        return FeffPot(structure, central_atom)

    def get_feff_atoms(self, structure, central_atom):
        """
        Creates string representation of atomic shell coordinates using in
        ATOMS file and feff.inp.

        Args:
            structure: pymatgen structure object
            central_atom: atom symbol string for absorbing atom

        Returns:
            FeffAtoms object
        """
        return FeffAtoms(structure, central_atom)

    def __str__(self):
        output = [self.name]
        section_names = ["XANES", "EXAFS"]
        for ns in section_names:
            for d in [self.xanes_settings, self.exafs_settings]:
                output.append(ns)
                for k, v in six.iteritems(d):
                    output.append("%s = %s" % (k, str(v)))
                output.append("")

        return "\n".join(output)


class MaterialsProjectFeffInputSet(FeffInputSet):
    """
    Implementation of FeffInputSet utilizing parameters in the public
    Materials Project.
    """
    def __init__(self):
        super(MaterialsProjectFeffInputSet, self).__init__("MaterialsProject")



# coding: utf-8

from __future__ import division, unicode_literals, print_function

"""
GW tasks for fireworks workflows, currently Vasp only
"""


__author__ = "Michiel van Setten"
__copyright__ = " "
__version__ = "0.9"
__maintainer__ = "Michiel van Setten"
__email__ = "mjvansetten@gmail.com"
__date__ = "May 2014"

import os
import os.path
import logging
import shutil
import subprocess
import socket

from pymatgen.core.structure import Structure
from pymatgen.io.gwwrapper.GWvaspinputsets import SingleVaspGWWork
from pymatgen.io.gwwrapper.GWvaspinputsets import GWG0W0VaspInputSet

try:
    from fireworks.core.firework import FireTaskBase, FWAction
    from fireworks.utilities.fw_serializers import FWSerializable
except ImportError:
    # Dummy classes 
    class FireTaskBase(object): pass
    class FWAction(object): pass
    class FWSerializable(object): pass


MODULE_DIR = os.path.dirname(os.path.abspath(__file__))

"""
MPGWVaspInputSet.joson contains the standards for GW calculations. This set contains all
parameters for the first sc dft calculation. The modifications for the subsequent
sub calculations are made below.
For many settings the number of cores on which the calculations will be run is needed, this
number is assumed to be on the environment variable NPARGWCALC.
"""


def get_mpi_runner_ntasks_keyword(mpi_runner):
    if mpi_runner == 'srun':
        return '-n'
    elif mpi_runner == 'mpirun':
        return '-np'
    else:
        raise EnvironmentError('Keyword for <number of tasks to run> is not found '
                               'for mpi_runner {}'.format(mpi_runner))


def get_vasp_environment(calculation_type='standard'):
    """
    Method to get the vasp environmental variables
    :param calculation_type: Type of calculation for the Vasp job ('standard' or 'hse' at the moment)
    :return: Dictionary with the necessary variables (vasp executable, vasp modules, vasp pseudos and mpi_runner)
    on a particular cluster for a particular type of calculation
    :raise ValueError: if the type of calculation is not allowed
    """
    if calculation_type == 'standard':
        try:
            vasp_exe = os.environ['VASP_EXE']
        except KeyError:
            hostname = socket.gethostname()
            raise EnvironmentError('VASP_EXE environment variable on '
                                   'cluster "{hn}" not found'.format(hn=hostname))
        try:
            vasp_modules = os.environ['VASP_MODULES'].split()
        except KeyError:
            hostname = socket.gethostname()
            raise EnvironmentError('VASP_MODULES environment variable on '
                                   'cluster "{hn}" not found'.format(hn=hostname))
        try:
            vasp_mpi_runner = os.environ['VASP_MPI_RUNNER']
        except KeyError:
            hostname = socket.gethostname()
            raise EnvironmentError('VASP_MPI_RUNNER environment variable on '
                                   'cluster "{hn}" not found'.format(hn=hostname))
        try:
            vasp_psp_dir = os.environ['VASP_PSP_DIR']
        except KeyError:
            hostname = socket.gethostname()
            raise EnvironmentError('VASP_PSP_DIR environment variable on '
                                   'cluster "{hn}" not found'.format(hn=hostname))
    elif calculation_type == 'hse':
        try:
            vasp_exe = os.environ['VASP_HSE_EXE']
        except KeyError:
            hostname = socket.gethostname()
            raise EnvironmentError('VASP_HSE_EXE environment variable on '
                                   'cluster "{hn}" not found'.format(hn=hostname))
        try:
            vasp_modules = os.environ['VASP_HSE_MODULES'].split()
        except KeyError:
            hostname = socket.gethostname()
            raise EnvironmentError('VASP_HSE_MODULES environment variable on '
                                   'cluster "{hn}" not found'.format(hn=hostname))
        try:
            vasp_mpi_runner = os.environ['VASP_HSE_MPI_RUNNER']
        except KeyError:
            hostname = socket.gethostname()
            raise EnvironmentError('VASP_HSE_MPI_RUNNER environment variable on '
                                   'cluster "{hn}" not found'.format(hn=hostname))
        try:
            vasp_psp_dir = os.environ['VASP_HSE_PSP_DIR']
        except KeyError:
            hostname = socket.gethostname()
            raise EnvironmentError('VASP_HSE_PSP_DIR environment variable on '
                                   'cluster "{hn}" not found'.format(hn=hostname))
    else:
        raise ValueError('Vasp calculation type "{}" is not allowed '
                         'in get_vasp_environment'.format(calculation_type))
    return {'vasp_exe': vasp_exe, 'vasp_modules': vasp_modules,
            'vasp_mpi_runner': vasp_mpi_runner, 'vasp_psp_dir': vasp_psp_dir,
            'mpi_runner_ntasks_option': get_mpi_runner_ntasks_keyword(vasp_mpi_runner)}


class VaspGWTask(FireTaskBase, FWSerializable):
    """
    Base class for vasp GW tasks
    """
    _fw_name = "Vasp GW task"

    def __init__(self, parameters):
        self.update(parameters)
        structure = Structure.from_dict(parameters['structure'])
        structure.vbm_l = parameters['band_structure']['vbm_l']
        structure.cbm_l = parameters['band_structure']['cbm_l']
        structure.vbm = (parameters['band_structure']['vbm_a'], parameters['band_structure']['vbm_b'], parameters['band_structure']['vbm_c'])
        structure.cbm = (parameters['band_structure']['cbm_a'], parameters['band_structure']['cbm_b'], parameters['band_structure']['cbm_c'])
        self.structure = structure
        self.job = parameters['job']
        self.spec = parameters['spec']
        self.option = parameters['option']

    def get_system(self):
        return self.structure.composition.reduced_formula

    def get_prep_dir(self):
        launch_dir = self.get_system()
        if self.option is not None:
            launch_dir = launch_dir + '.' + self.option['test_prep'] + str(self.option['value_prep'])
        return launch_dir

    def get_launch_dir(self):
        launch_dir = self.get_prep_dir()
        if self.job not in 'prep':
            if 'test' in self.option.keys():
                option_name = '.' + self.option['test'] + str(self.option['value'])
            else:
                option_name = ''
            launch_dir = os.path.join(launch_dir, self.job + option_name)
        return launch_dir


class VaspGWInputTask(VaspGWTask):
    _fw_name = "Vasp GW Input Task"
    """
    class for creating input for GW calculations
    """
    def __init__(self, parameters):
        VaspGWTask.__init__(self, parameters)

    def run_task(self, fw_spec):
        task = SingleVaspGWWork(structure=self.structure, job=self.job, spec=self.spec, option=self.option)
        task.create_input()
        return FWAction()


class VaspGWToDiagTask(VaspGWTask):
    _fw_name = 'Vasp GW To Diag Task'
    """
    class for copying the INCAR.DIAG to INCAR
    """
    def __init__(self, parameters):
        VaspGWTask.__init__(self, parameters)

    def run_task(self, fw_spec):
        shutil.copy2(os.path.join(self.get_prep_dir(), 'INCAR.DIAG'), os.path.join(self.get_prep_dir(), 'INCAR'))
        return FWAction()


class VaspGWGetPrepResTask(VaspGWTask):
    _fw_name = "Vasp GW Get Prep Res Task"
    """
    class for copying the results of the preparation job, the dft calculation, to the GW work folder
    """
    def __init__(self, parameters):
        VaspGWTask.__init__(self, parameters)

    def run_task(self, fw_spec):
        output_files = ['CHGCAR', 'WAVECAR', 'WAVEDER']
        for output_file in output_files:
            shutil.copy2(os.path.join(self.get_prep_dir(), output_file), self.get_launch_dir())
        return FWAction()


class VaspGWExecuteTask(VaspGWTask):
    _fw_name = 'Vasp GW Execute Task'
    """
    class for executing vasp
    """
    def __init__(self, parameters):
        VaspGWTask.__init__(self, parameters)

    def run_task(self, fw_spec):
        name = self.get_system() + self.job
        logging.basicConfig(filename=os.path.join(self.get_launch_dir(), name + '.log'), level=logging.DEBUG)
        vasp_exe = get_vasp_environment()['vasp_exe']
        vasp_mpi_executer = get_vasp_environment()['vasp_mpi_executer']
        n_tasks = GWG0W0VaspInputSet(self.structure).get_npar(self.structure)

        frontend_serial = True
        custodian = False

        if frontend_serial:
            #fake run, no actual vasp execution
            base = os.getcwdu()
            abs_work_dir = os.path.join(base, self.get_launch_dir())
            os.chdir(abs_work_dir)
            cmd = ["/home/setten/scripts/vasp"]
            subprocess.call(cmd)
            os.chdir(base)
        elif custodian:
            mpirunvasp = [vasp_mpi_executer, '-np', n_tasks, vasp_exe]
            pass

        return FWAction()


class VaspGWWriteConDatTask(VaspGWTask):
    fw_name = "Vasp GW Write Convergence Data"
    """
    Write the data needed to study the convergence

    """
    def __init__(self, parameters):
        VaspGWTask.__init__(self, parameters)

    def run_task(self, fw_spec):
        return FWAction()


class VaspGWTestConTask(VaspGWTask):
    fw_name = "Vasp GW Test converge"
    """
    test if the current conv option is converged if not calculate the nex value

    """
    def __init__(self, parameters):
        VaspGWTask.__init__(self, parameters)

    def get_n_pev_runs(self):
        """
        return the number of values that have already been calculated for this convergence parameter
        """
        n = 0
        return n

    def run_task(self, fw_spec):
        task = SingleVaspGWWork(structure=self.structure, job=self.job, spec=self.spec, option=self.option)
        task.create_input()
        return FWAction()



from __future__ import unicode_literals

"""
This package contains core modules and classes for representing structures and
operations on them.
"""

__author__ = "Shyue Ping Ong"
__date__ = "Dec 15, 2010 7:21:29 PM"

from .periodic_table import Element, Specie, DummySpecie, get_el_sp
from .composition import Composition, ChemicalPotential
from .structure import Structure, IStructure, Molecule, IMolecule
from .bonds import CovalentBond, get_bond_length
from .lattice import Lattice
from .sites import Site, PeriodicSite
from .operations import SymmOp
from .units import *
